from __future__ import annotations
from typing import Literal, Any
import matplotlib
import matplotlib.pyplot as plt
import astropy.units as u
import numpy as np
import numpy.typing as npt
import named_arrays as na

__all__ = [
    "subplots",
    "plot",
    "fill",
    "scatter",
    "imshow",
    "pcolormesh",
    "text",
    "brace_vertical",
]


def subplots(
        axis_rows: str = "subplots_row",
        ncols: int = 1,
        axis_cols: str = "subplots_col",
        nrows: int = 1,
        *,
        sharex: bool | Literal["none", "all", "row", "col"] = False,
        sharey: bool | Literal["none", "all", "row", "col"] = False,
        squeeze: bool = True,
        **kwargs,
) -> tuple[
    matplotlib.figure.Figure,
    matplotlib.axes.Axes | na.ScalarArray[npt.NDArray],
]:
    """
    A thin wrapper around :func:`matplotlib.pyplot.subplots()` which allows for
    providing axis names to the rows and columns.

    Parameters
    ----------
    axis_rows
        Name of the axis representing the rows in the subplot grid.
        If :obj:`None`, the ``squeeze`` argument must be :obj:`True`.
    nrows
        Number of rows in the subplot grid
    axis_cols
        Name of the axis representing the columns in the subplot grid
        If :obj:`None`, the ``squeeze`` argument must be :obj:`True`.
    ncols
        Number of columns in the subplot grid
    sharex
        Controls whether all the :class:`matplotlib.axes.Axes` instances share the same horizontal axis properties.
        See the documentation of :func:`matplotlib.pyplot.subplots` for more information.
    sharey
        Controls whether all the :class:`matplotlib.axes.Axes` instances share the same vertical axis properties.
        See the documentation of :func:`matplotlib.pyplot.subplots` for more information.
    squeeze
        If :obj:`True`, :func:`numpy.squeeze` is called on the result, which removes singleton dimensions from the
        array.
        See the documentation of :func:`matplotlib.pyplot.subplots` for more information.
    kwargs
        Additional keyword arguments passed to :func:`matplotlib.pyplot.subplots`
    """

    shape = {axis_rows: nrows, axis_cols: ncols}

    if squeeze:
        shape = {axis: shape[axis] for axis in shape if shape[axis] != 1}

    fig, axs = plt.subplots(
        ncols=ncols,
        nrows=nrows,
        sharex=sharex,
        sharey=sharey,
        squeeze=squeeze,
        **kwargs,
    )

    return fig, na.ScalarArray(axs, axes=tuple(shape.keys()))


def plot(
        *args: na.AbstractScalar,
        ax: None | matplotlib.axes.Axes | na.ScalarArray[npt.NDArray[matplotlib.axes.Axes]] = None,
        axis: None | str = None,
        where: bool | na.AbstractScalar = True,
        transformation: None | na.transformations.AbstractTransformation = None,
        components: None | tuple[str, ...] = None,
        **kwargs,
) -> na.ScalarArray[npt.NDArray[None | matplotlib.artist.Artist]]:
    """
    A thin wrapper around :meth:`matplotlib.axes.Axes.plot` for named arrays.

    The main difference of this function from :func:`matplotlib.pyplot.plot` is the addition of the ``axis`` parameter
    indicating along which axis the lines should be connected.

    Parameters
    ----------
    args
        Same signature as :meth:`matplotlib.axes.Axes.plot`.
        If ``ax`` is a 2D plot, ``*args`` should be ``y`` or ``x, y``.
        If ``ax`` is a 3D plot, ``*args`` should be ``x, y`` or ``x, y, z``.
    ax
        The instances of :class:`matplotlib.axes.Axes` to use.
        If :obj:`None`, calls :func:`matplotlib.pyplot.gca` to get the current axes.
        If an instance of :class:`named_arrays.ScalarArray`, ``ax.shape`` should be a subset of the broadcasted shape of
        ``*args``.
    axis
        The name of the axis that the plot lines should be connected along.
        If :obj:`None`, the broadcasted shape of ``args`` should have only one element,
        otherwise a :class:`ValueError` is raised.
    where
        A boolean array that selects which elements to plot
    transformation
        A callable that is applied to args before plotting
    components
        The component names of ``*args`` to plot, helpful if ``*args`` are an instance of
        :class:`named_arrays.AbstractVectorArray`.
    kwargs
        Additional keyword arguments passed to :meth:`matplotlib.axes.Axes.plot`.
        These can be instances of :class:`named_arrays.AbstractArray`.

    Returns
    -------
        An array of artists that were plotted

    Examples
    --------

    Plot a single scalar

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import named_arrays as na

        x = na.linspace(0, 2 * np.pi, axis="x",  num=101)
        y = np.sin(x)

        plt.figure();
        na.plt.plot(x, y);

    Plot an array of scalars

    .. jupyter-execute::

        z = na.linspace(0, np.pi, axis="z", num=5)

        y = np.sin(x - z)

        plt.figure();
        na.plt.plot(x, y, axis="x");

    Plot an uncertain scalar

    .. jupyter-execute::

        ux = na.NormalUncertainScalarArray(x, width=0.2)
        uy = np.sin(ux)

        plt.figure();
        na.plt.plot(x, uy);

    Broadcast an array of scalars against an array of :class:`matplotlib.axes.Axes`

    .. jupyter-execute::

        fig, ax = na.plt.subplots(axis_rows="z", nrows=z.shape["z"], sharex=True)

        na.plt.plot(x, y, ax=ax, axis="x");

    Plot a 2D Cartesian vector

    .. jupyter-execute::

        v = na.Cartesian2dVectorArray(x, np.sin(x))

        plt.figure()
        na.plt.plot(v)

    """
    if transformation is not None:
        args = tuple(transformation(arg) for arg in args)
    return na._named_array_function(
        plot,
        *args,
        ax=ax,
        axis=axis,
        where=where,
        components=components,
        **kwargs,
    )


def fill(
        *args: na.AbstractArray,
        ax: None | matplotlib.axes.Axes | na.ScalarArray[npt.NDArray] = None,
        axis: None | str = None,
        where: bool | na.AbstractScalar = True,
        transformation: None | na.transformations.AbstractTransformation = None,
        components: None | tuple[str, ...] = None,
        **kwargs,
) -> na.ScalarArray[npt.NDArray]:
    """
    Plot filled polygons

    This is a thin wrapper around :meth:`matplotlib.axes.Axes.fill` for named arrays.

    The main difference of this function from :func:`matplotlib.pyplot.fill` is the addition of the ``axis`` parameter
    indicating along which axis the lines should be connected.

    Parameters
    ----------
    args
        Same signature as :meth:`matplotlib.axes.Axes.fill`.
        If ``ax`` is a 2D plot, ``*args`` should be ``x, y``.
        If ``ax`` is a 3D plot, ``*args`` should be ``x, y, z``.
    ax
        The instances of :class:`matplotlib.axes.Axes` to use.
        If :obj:`None`, calls :func:`matplotlib.pyplot.gca` to get the current axes.
        If an instance of :class:`named_arrays.ScalarArray`, ``ax.shape`` should be a subset of the broadcasted shape of
        ``*args``.
    axis
        The name of the axis that the plot lines should be connected along.
        If :obj:`None`, the broadcasted shape of ``args`` should have only one element,
        otherwise a :class:`ValueError` is raised.
    where
        A boolean array that selects which elements to plot
    transformation
        A callable that is applied to args before plotting
    components
        The component names of ``*args`` to plot, helpful if ``*args`` are an instance of
        :class:`named_arrays.AbstractVectorArray`.
    kwargs
        Additional keyword arguments passed to :meth:`matplotlib.axes.Axes.fill`.
        These can be instances of :class:`named_arrays.AbstractArray`.
    """
    if transformation is not None:
        args = tuple(transformation(arg) for arg in args)
    return na._named_array_function(
        fill,
        *args,
        ax=ax,
        axis=axis,
        where=where,
        components=components,
        **kwargs,
    )


def scatter(
        *args: na.AbstractScalar,
        s: None | na.AbstractScalarArray = None,
        c: None | na.AbstractScalarArray = None,
        ax: None | matplotlib.axes.Axes | na.ScalarArray[npt.NDArray[matplotlib.axes.Axes]] = None,
        where: bool | na.AbstractScalar = True,
        transformation: None | na.transformations.AbstractTransformation = None,
        components: None | tuple[str, ...] = None,
        **kwargs,
) -> na.ScalarArray[npt.NDArray[None | matplotlib.artist.Artist]]:
    """
    A thin wrapper around :meth:`matplotlib.axes.Axes.scatter` for named arrays.

    Parameters
    ----------
    args
        Same signature as :meth:`matplotlib.axes.Axes.scatter`.
        If ``ax`` is a 2D plot, ``*args`` should be ``y`` or ``x, y``.
        If ``ax`` is a 3D plot, ``*args`` should be ``x, y`` or ``x, y, z``.
    s
        The marker size in points**2
    c
        The color of the markers
    ax
        The instances of :class:`matplotlib.axes.Axes` to use.
        If :obj:`None`, calls :func:`matplotlib.pyplot.gca` to get the current axes.
        If an instance of :class:`named_arrays.ScalarArray`, ``ax.shape`` should be a subset of the broadcasted shape of
        ``*args``.
    where
        A boolean array that selects which elements to plot
    transformation
        A callable that is applied to args before plotting
    components
        The component names of ``*args`` to plot, helpful if ``*args`` are an instance of
        :class:`named_arrays.AbstractVectorArray`.
    kwargs
        Additional keyword arguments passed to :meth:`matplotlib.axes.Axes.scatter`.
        These can be instances of :class:`named_arrays.AbstractArray`.

    Examples
    --------

    Plot a single scalar

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import named_arrays as na

        x = na.linspace(0, 2 * np.pi, axis="x",  num=51)
        y = np.sin(x)

        plt.figure();
        na.plt.scatter(x, y);

    Plot an array of scalars

    .. jupyter-execute::

        z = na.linspace(0, np.pi, axis="z", num=5)

        y = np.sin(x - z)

        plt.figure();
        na.plt.scatter(x, y, c=z);

    Plot an uncertain scalar

    .. jupyter-execute::

        ux = na.NormalUncertainScalarArray(x, width=0.2)
        uy = np.sin(ux)

        plt.figure();
        na.plt.scatter(x, uy);

    Broadcast an array of scalars against an array of :class:`matplotlib.axes.Axes`

    .. jupyter-execute::

        fig, ax = na.plt.subplots(axis_rows="z", nrows=z.shape["z"], sharex=True)

        na.plt.scatter(x, y, ax=ax);
    """
    if transformation is not None:
        args = tuple(transformation(arg) for arg in args)
    return na._named_array_function(
        scatter,
        *args,
        s=s,
        c=c,
        ax=ax,
        where=where,
        components=components,
        **kwargs,
    )


def imshow(
    X: na.AbstractArray,
    *,
    axis_x: str,
    axis_y: str,
    axis_rgb: None | str = None,
    ax: None | matplotlib.axes.Axes | na.AbstractArray = None,
    cmap: None | str | matplotlib.colors.Colormap = None,
    norm: None | str | matplotlib.colors.Normalize = None,
    aspect: None | na.ArrayLike = None,
    alpha: None | na.ArrayLike = None,
    vmin: None | na.ArrayLike = None,
    vmax: None | na.ArrayLike = None,
    extent: None | na.ArrayLike = None,
    **kwargs,
) -> na.AbstractArray:
    """
    A thin wrappper around :func:`matplotlib.pyplot.imshow` for named arrays.

    Parameters
    ----------
    X
        The image data.
    axis_x
        The name of the horizontal axis.
    axis_y
        The name of the vertical axis
    axis_rgb
        The optional name of the color axis.
    ax
        The instances of :class:`matplotlib.axes.Axes` to use.
        If :obj:`None`, calls :func:`matplotlib.pyplot.gca` to get the current axes.
        If an instance of :class:`named_arrays.ScalarArray`, ``ax.shape`` should be a subset of the broadcasted shape of
        ``*args``.
    cmap
        The Colormap instance or registered colormap name used to map scalar
        data to colors.
    norm
        The normalization method used to scale the data into the range 0 to 1
        before mapping colors.
    aspect
        The aspect ratio of the Axes.
    alpha
        The alpha blending value that ranges between 0 (transparent) and 1 (opaque).
    vmin
        The minimum value of the data range.
    vmax
        The maximum value of the data range.
    extent
        The bounding box in data coordinates that the image will fill.
        The logical axis name of the bounding box should be ``f"{axis_x},{axis_y}``
    kwargs
        An additional keyword arguments that are passed to :func:`matplotlib.pyplot.imshow`.

    Examples
    --------

    Plot a random 2d array.

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import named_arrays as na

        # Define the horizontal and vertical axis names of the image
        axis_x = "x"
        axis_y = "y"

        # Define a random dummy array to plot
        a = na.random.uniform(
            low=0,
            high=1,
            shape_random={axis_x: 16, axis_y: 16},
        )

        # Create the plot axes
        fig, ax = plt.subplots()

        # Plot the dummy array
        na.plt.imshow(
            a,
            axis_x=axis_x,
            axis_y=axis_y,
            ax=ax,
            extent=na.ScalarArray(
                ndarray=np.array([0, 1, 0, 1]),
                axes=f"{axis_x},{axis_y}",
            ),
        );

    |

    Plot a grid of 2d dummy arrays

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import named_arrays as na

        # Define the horizontal and vertical axis names of the plot
        axis_x = "x"
        axis_y = "y"

        # Define the row and column axis names of the plot
        axis_rows = "row"
        axis_cols = "col"

        # Define the number of rows and columns of the plot
        num_rows = 2
        num_cols = 3

        # Define a random dummy array to plot
        a = na.random.uniform(
            low=0,
            high=1,
            shape_random={
                axis_rows: num_rows,
                axis_cols: num_cols,
                axis_x: 16,
                axis_y: 16,
            }
        )

        # Define the array of matplotlib axes
        fig, axs = na.plt.subplots(
            axis_rows=axis_rows,
            nrows=num_rows,
            axis_cols=axis_cols,
            ncols=num_cols,
            sharex=True,
            sharey=True,
            constrained_layout=True,
        )

        # Plot the dummy arrays
        na.plt.imshow(
            a,
            axis_x=axis_x,
            axis_y=axis_y,
            ax=axs,
            extent=na.ScalarArray(
                ndarray=np.array([0, 1, 0, 1]),
                axes=f"{axis_x},{axis_y}",
            ),
        );
    """
    return na._named_array_function(
        func=imshow,
        X=X,
        axis_x=axis_x,
        axis_y=axis_y,
        axis_rgb=axis_rgb,
        ax=ax,
        cmap=cmap,
        norm=norm,
        aspect=aspect,
        alpha=alpha,
        vmin=vmin,
        vmax=vmax,
        extent=extent,
        **kwargs,
    )


def pcolormesh(
    *XY: na.AbstractArray,
    C: na.AbstractArray,
    components: None | tuple[str, str] = None,
    axis_rgb: None | str = None,
    ax: None | matplotlib.axes.Axes | na.AbstractArray = None,
    cmap: None | str | matplotlib.colors.Colormap = None,
    norm: None | str | matplotlib.colors.Normalize = None,
    vmin: None | na.ArrayLike = None,
    vmax: None | na.ArrayLike = None,
    **kwargs,
) -> na.AbstractScalar:
    """
    A thin wrapper around :func:`matplotlib.pyplot.pcolormesh` for named arrays.

    Parameters
    ----------
    XY
        The coordinates of the mesh.
        If `C` is a scalar, `XY` can either be two scalars or one vector .
        If `C` is a function, `XY` is not specified.
        If `XY` is not specified as two scalars, the `components` must be given,
        see below.
    C
        The mesh data.
    components
        If `XY` is not specified as two scalars, this parameter should
        be a tuple of two strings, specifying the vector components of `XY`
        to use as the horizontal and vertical components of the mesh.
    axis_rgb
        The optional logical axis along which the RGB color channels are
        distributed.
    ax
        The instances of :class:`matplotlib.axes.Axes` to use.
        If :obj:`None`, calls :func:`matplotlib.pyplot.gca` to get the current axes.
        If an instance of :class:`named_arrays.ScalarArray`, ``ax.shape`` should be a subset of the broadcasted shape of
        ``*args``.
    cmap
        The colormap used to map scalar data to colors.
    norm
        The normalization method used to scale data into the range [0, 1] before
        mapping to colors.
    vmin
        The minimum value of the data range.
    vmax
        The maximum value of the data range.
    kwargs
        Additional keyword arguments accepted by `matplotlib.pyplot.pcolormesh`

    Examples
    --------

    Plot a random 2D mesh

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import named_arrays as na

        # Define the size of the grid
        shape = dict(x=16, y=16)

        # Define a simple coordinate grid
        x = na.linspace(-2, 2, axis="x", num=shape["x"])
        y = na.linspace(-1, 1, axis="y", num=shape["y"])

        # Define a random 2D array of values to plot
        a = na.random.uniform(-1, 1, shape_random=shape)

        # Plot the coordinates and values using pcolormesh
        fig, ax = plt.subplots(constrained_layout=True)
        na.plt.pcolormesh(x, y, C=a, ax=ax);

    |

    Plot a grid of random 2D meshes

    .. jupyter-execute::

        import named_arrays as na

        # Define the size of the grid
        shape = dict(row=2, col=3, x=16, y=16)

        # Define a simple coordinate grid
        x = na.linspace(-2, 2, axis="x", num=shape["x"])
        y = na.linspace(-1, 1, axis="y", num=shape["y"])

        # Define a random 2D array of values to plot
        a = na.random.uniform(-1, 1, shape_random=shape)

        # Plot the coordinates and values using pcolormesh
        fig, ax = na.plt.subplots(
            axis_rows="row",
            nrows=shape["row"],
            axis_cols="col",
            ncols=shape["col"],
            sharex=True,
            sharey=True,
            constrained_layout=True,
        )
        na.plt.pcolormesh(x, y, C=a, ax=ax);
    """
    return na._named_array_function(
        pcolormesh,
        *XY,
        C=C,
        axis_rgb=axis_rgb,
        ax=ax,
        cmap=cmap,
        norm=norm,
        vmin=vmin,
        vmax=vmax,
        components=components,
        **kwargs,
    )


def text(
    x: float | u.Quantity | na.AbstractScalar,
    y: float | u.Quantity | na.AbstractScalar,
    s: str | na.AbstractScalar,
    ax: None | matplotlib.axes.Axes | na.AbstractScalar = None,
    **kwargs,
) -> na.AbstractScalar:
    """
    A thin wrapper around :meth:`matplotlib.axes.Axes.text` for named arrays.

    Parameters
    ----------
    x
        The horizontal position of the text in data coordinates.
    y
        The vertical position of the text in data coordinates.
    s
        The text to plot.
    ax
        The matplotlib axes instance on which to plot the text.
    kwargs
        Additional keyword arguments to pass to :meth:`matplotlib.axes.Axes.text`.

    Examples
    --------

    Plot an array of text values at different locations.

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u
        import named_arrays as na

        # Create an array of azimuths where the next will be placed
        azimuth = na.linspace(0, 360, axis="azimuth", num=4, endpoint=False) * u.deg

        # Compute the x and y coordinates for the given azimuth
        x = np.cos(azimuth)
        y = np.sin(azimuth)

        # Create an array of strings to plot at the chosen positions
        s = na.ScalarArray(
            ndarray=np.array(["East", "North", "West", "South"]),
            axes="azimuth"
        )

        # Plot the array of strings
        fig, ax = plt.subplots()
        na.plt.text(
            x=x,
            y=y,
            s=s,
            ax=ax,
        );
        ax.set_xlim(-2, 2);
        ax.set_ylim(-2, 2);
    """
    return na._named_array_function(
        text,
        x=x,
        y=y,
        s=s,
        ax=ax,
        **kwargs,
    )


def brace_vertical(
    x: float | u.Quantity | na.AbstractScalar,
    width: float | u.Quantity | na.AbstractScalar,
    ymin: float | u.Quantity | na.AbstractScalar,
    ymax: float | u.Quantity | na.AbstractScalar,
    ax: None | matplotlib.axes | na.AbstractScalar = None,
    label: None | str | na.AbstractScalar = None,
    beta: None | float | na.AbstractScalar = None,
    kind: Literal["left", "right"] = "left",
    kwargs_plot: None | dict[str, Any] = None,
    kwargs_text: None | dict[str, Any] = None,
    **kwargs,
) -> na.AbstractScalar:
    """
    Plot a vertical curly bracket at the given coordinates.

    Parameters
    ----------
    x
        The horizontal position of the vertical curly bracket.
    width
        The width of the curly bracket in data coordinates.
    ymin
        The minimum span of the vertical curly bracket.
    ymax
        The maximum span of the vertical curly bracket.
    ax
        A matplotlib axes instance on which to plot the curly bracket.
    label
        The optional text label for the curly bracket.
    beta
        Parameter which controls the "curlyness" of the bracket.
        If :obj:`None`, ``beta = 2 / width``.
    kind
        The kind of the brace, left or right.
    kwargs_plot
        Additional keyword arguments that are passed to :func:`plot`.
    kwargs_text
        Additional keyword arguments that are passed to :func:`text`.
    kwargs
        Additional keyword arguments that are passed to both
        :func:`plot` and :func:`text`.

    Examples
    --------

    Plot an array of braces with different lengths

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import named_arrays as na

        # Define the number of braces to plot
        num = 5

        # Define the x coordinate of the braces
        x = na.linspace(0.2, 0.9, axis="y", num=num)

        # Define the y ranges of the braces
        ymin = -na.linspace(.3, .8, axis="y", num=num)
        ymax = +na.linspace(.3, .8, axis="y", num=num)

        # Define the label as the length of the brace
        label = ymax - ymin

        # Plot the braces
        fig, ax = plt.subplots()
        na.plt.brace_vertical(
            x=x,
            width=0.05,
            ymin=ymin,
            ymax=ymax,
            ax=ax,
            kind="left",
            label=label,
        );
        ax.set_xlim(0, 1);
        ax.set_ylim(-1, 1);
    """
    if kwargs_plot is None:
        kwargs_plot = dict()
    kwargs_plot = kwargs | kwargs_plot

    if kwargs_text is None:
        kwargs_text = dict()
    kwargs_text = kwargs | kwargs_text

    label = na.as_named_array(label).astype(str).astype(object)

    if beta is None:
        beta = 1 / (width / 2)

    axis = "_brace"

    y = na.linspace(ymin, ymax, axis=axis, num=1001)

    ycen = (ymin + ymax) / 2
    z = np.abs(y - ycen)

    f_outer = 1 / (1 + np.exp(-beta * (z - z.min(axis))))
    f_inner = 1 / (1 + np.exp(-beta * (z - z.max(axis))))

    f = f_outer + f_inner

    f = f - 1.5

    if kind == "left":
        x_text = x - width
        ha = "right"
        label = label + "  "
    elif kind == "right":
        f = -f
        x_text = x + width
        ha = "left"
        label = "  " + label
    else:   # pragma: nocover
        raise ValueError(
            f"Invalide kind of brace '{kind}', the only supported options are "
            f"'left' and 'right'."
        )

    f = x + width * f

    result = plot(
        f,
        y,
        ax=ax,
        axis=axis,
        **kwargs_plot,
    )

    text(
        x=x_text,
        y=ycen,
        s=label,
        ax=ax,
        ha=ha,
        va="center",
        **kwargs_text
    )

    return result



from typing import Sequence, Callable
import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
import named_arrays.tests.test_core
import named_arrays._vectors.cartesian.tests.test_vectors_cartesian_2d

__all__ = [
    "AbstractTestAbstractFunctionArray",
]

_num_x = named_arrays.tests.test_core.num_x
_num_y = named_arrays.tests.test_core.num_y
_num_distribution = named_arrays.tests.test_core.num_distribution


def _function_arrays():
    functions_1d = [
        na.FunctionArray(
            inputs=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y),
            outputs=na.ScalarUniformRandomSample(-5, 5, shape_random=dict(y=_num_y))
        )
    ]

    inputs_2d = [
        na.ScalarLinearSpace(0, 1, axis='y', num=_num_y),
        na.Cartesian2dVectorLinearSpace(
            start=0,
            stop=1,
            axis=na.Cartesian2dVectorArray('x', 'y'),
            num=na.Cartesian2dVectorArray(_num_x, _num_y)
        )
    ]

    outputs_2d = [
        na.ScalarUniformRandomSample(-5, 5, shape_random=dict(x=_num_x, y=_num_y)),
        na.UniformUncertainScalarArray(
            nominal=na.ScalarUniformRandomSample(-5, 5, shape_random=dict(x=_num_x, y=_num_y)),
            width=1,
            num_distribution=_num_distribution,
        ),
        na.Cartesian2dVectorUniformRandomSample(-5, 5, shape_random=dict(x=_num_x, y=_num_y)),
    ]

    functions_2d = [
        na.FunctionArray(
            inputs=inputs,
            outputs=outputs,
        )
        for inputs in inputs_2d
        for outputs in outputs_2d
    ]

    functions = functions_1d + functions_2d

    return functions


def _function_arrays_2():
    return (
        6,
        na.ScalarArray(6),
        na.UniformUncertainScalarArray(6, width=1, num_distribution=_num_distribution),
        na.Cartesian2dVectorArray(6, 7),
        na.FunctionArray(
            inputs=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y),
            outputs=na.ScalarUniformRandomSample(-6, 6, shape_random=dict(y=_num_y))
        ),
        na.FunctionArray(
            inputs=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y),
            outputs=na.Cartesian2dVectorUniformRandomSample(-6, 6, shape_random=dict(y=_num_y))
        )
    )


class AbstractTestAbstractFunctionArray(
    named_arrays.tests.test_core.AbstractTestAbstractArray,
):

    def test_inputs(self, array: na.AbstractFunctionArray):
        assert isinstance(array.inputs, na.AbstractArray)

    def test_outputs(self, array: na.AbstractFunctionArray):
        assert isinstance(array.outputs, na.AbstractArray)

    @pytest.mark.parametrize("dtype", (int, float))
    def test_astype(self, array: na.AbstractFunctionArray, dtype: type):
        super().test_astype(array=array, dtype=dtype)
        result = array.astype(dtype)
        if isinstance(result.outputs, na.AbstractScalar):
            assert result.outputs.dtype == dtype
        elif isinstance(result.outputs, na.AbstractVectorArray):
            for e in result.outputs.entries:
                assert result.outputs.entries[e].dtype == dtype

    @pytest.mark.parametrize('unit', [u.m, u.s])
    def test_to(self, array: na.AbstractFunctionArray, unit: None | u.UnitBase):
        super().test_to(array=array, unit=unit)
        if isinstance(array.outputs, na.ScalarArray):
            if na.unit_normalized(array.outputs).is_equivalent(unit):
                result = array.to(unit)
                assert result.outputs.unit == unit
                return

        if isinstance(array.outputs, na.AbstractVectorArray):
            entries = array.outputs.entries
            if all(unit.is_equivalent(na.unit_normalized(entries[e])) for e in entries):
                result = array.to(unit)
                assert result.type_abstract == array.type_abstract
                assert all(result.outputs.entries[e].unit == unit for e in result.outputs.entries)
                return

        with pytest.raises(u.UnitConversionError):
            array.to(unit)

    def test_length(self, array: na.AbstractFunctionArray):
        result = array.length
        assert np.all(result.outputs.length == array.outputs.length)
        assert np.all(result.inputs == array.inputs)

    @pytest.mark.parametrize(
        argnames='item',
        argvalues=[
            dict(y=0),
            dict(y=slice(0, 1)),
            dict(y=na.ScalarArrayRange(0, 2, axis='y')),
            dict(
                y=na.FunctionArray(
                    inputs=na.ScalarArrayRange(0, 2, axis='y'),
                    outputs=na.ScalarArrayRange(0, 2, axis='y'),
                )
            ),
            na.FunctionArray(
                inputs=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y),
                outputs=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y) > 0.5,
            ),
            na.FunctionArray(
                inputs=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y),
                outputs=na.UniformUncertainScalarArray(
                    nominal=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y),
                    width=0.1,
                    num_distribution=_num_distribution,
                ) > 0.5,
            )
        ],
    )
    def test__getitem__(
            self,
            array: na.AbstractFunctionArray,
            item: dict[str, int | slice | na.AbstractArray] | na.AbstractArray
    ):
        super().test__getitem__(array=array, item=item)

        if not array.shape:
            with pytest.raises(ValueError):
                array[item]
            return

        if isinstance(item, na.AbstractFunctionArray):
            if np.any(item.inputs != array.inputs):
                with pytest.raises(ValueError):
                    array[item]
                return

            item_outputs = item_inputs = item.outputs

        elif isinstance(item, dict):
            item_inputs = {
                ax: item[ax].inputs if isinstance(item[ax], na.AbstractFunctionArray) else item[ax]
                for ax in item
            }
            item_outputs = {
                ax: item[ax].outputs if isinstance(item[ax], na.AbstractFunctionArray) else item[ax]
                for ax in item
            }

        result = array[item]

        assert np.all(result.inputs == array.inputs[item_inputs])
        assert np.all(result.outputs == array.outputs[item_outputs])

    def test__bool__(self, array: na.AbstractFunctionArray):
        if array.shape or array.unit is not None:
            with pytest.raises(
                expected_exception=ValueError,
                match=r"(Quantity truthiness is ambiguous, .*)"
                      r"|(The truth value of an array with more than one element is ambiguous. .*)"
            ):
                bool(array)
            return

        result = bool(array)
        assert isinstance(result, bool)

    def test__mul__(self, array: na.AbstractFunctionArray):
        unit = u.mm
        result = array * unit
        assert np.all(result.outputs == array.outputs * unit)
        assert np.all(result.inputs == array.inputs)

    def test__lshift__(self, array: na.AbstractFunctionArray):
        unit = u.mm
        result = array << unit
        assert np.all(result.outputs == array.outputs << unit)
        assert np.all(result.inputs == array.inputs)

    def test__truediv__(self, array: na.AbstractFunctionArray):
        unit = u.mm
        result = array / unit
        assert np.all(result.outputs == array.outputs / unit)
        assert np.all(result.inputs == array.inputs)

    class TestUfuncUnary(
        named_arrays.tests.test_core.AbstractTestAbstractArray.TestUfuncUnary
    ):

        def test_ufunc_unary(
                self,
                ufunc: np.ufunc,
                array: na.AbstractFunctionArray,
        ):
            try:
                outputs_expected = ufunc(array.outputs)
                inputs_expected = array.inputs
                if ufunc.nout != 1:
                    inputs_expected = (inputs_expected, ) * ufunc.nout
            except Exception as e:
                with pytest.raises(type(e)):
                    ufunc(array)
                return

            result = ufunc(array)

            if ufunc.nout == 1:
                out = 0 * result
            else:
                out = tuple(0 * r for r in result)

            result_out = ufunc(array, out=out)

            if ufunc.nout == 1:
                out = (out, )
                outputs_expected = (outputs_expected, )
                inputs_expected = (inputs_expected, )
                result = (result, )
                result_out = (result_out, )

            for i in range(ufunc.nout):
                assert np.all(result[i].outputs == outputs_expected[i])
                assert np.all(result[i].inputs == inputs_expected[i])
                assert np.all(result[i] == result_out[i])
                assert result_out[i] is out[i]

    @pytest.mark.parametrize("array_2", _function_arrays_2())
    class TestUfuncBinary(
        named_arrays.tests.test_core.AbstractTestAbstractArray.TestUfuncBinary
    ):

        def test_ufunc_binary(
                self,
                ufunc: np.ufunc,
                array: None | bool | int | float | complex | str | na.AbstractArray,
                array_2: None | bool | int | float | complex | str | na.AbstractArray,
        ):
            if isinstance(array, na.AbstractFunctionArray):
                array_normalized = array
                if isinstance(array_2, na.AbstractFunctionArray):
                    array_2_normalized = array_2
                else:
                    array_2_normalized = na.FunctionArray(inputs=array.inputs, outputs=array_2)
            else:
                if isinstance(array_2, na.AbstractFunctionArray):
                    array_normalized = na.FunctionArray(inputs=array_2.inputs, outputs=array)
                    array_2_normalized = array_2
                else:
                    raise TypeError("neither `array` nor `array_2` are instances of`named_arrays.FunctionArray`")

            try:
                outputs_expected = ufunc(array_normalized.outputs, array_2_normalized.outputs)
                inputs_expected = array_normalized.inputs
                if ufunc.nout != 1:
                    inputs_expected = (inputs_expected, ) * ufunc.nout
            except Exception as e:
                with pytest.raises(type(e)):
                    ufunc(array, array_2)
                return

            result = ufunc(array, array_2)

            if ufunc.nout == 1:
                out = 0 * result
            else:
                out = tuple(0 * r for r in result)

            result_out = ufunc(array, array_2, out=out)

            if ufunc.nout == 1:
                out = (out, )
                outputs_expected = (outputs_expected, )
                inputs_expected = (inputs_expected, )
                result = (result, )
                result_out = (result_out, )

            for i in range(ufunc.nout):
                assert np.all(result[i].outputs == outputs_expected[i])
                assert np.all(result[i].inputs == inputs_expected[i])
                assert np.all(result[i] == result_out[i])
                assert result_out[i] is out[i]

    @pytest.mark.parametrize("array_2", _function_arrays_2())
    class TestMatmul(
        named_arrays.tests.test_core.AbstractTestAbstractArray.TestMatmul
    ):

        def test_matmul(
                self,
                array: None | bool | int | float | complex | str | na.AbstractArray,
                array_2: None | bool | int | float | complex | str | na.AbstractArray,
        ):
            if isinstance(array, na.AbstractFunctionArray):
                array_normalized = array
                if isinstance(array_2, na.AbstractFunctionArray):
                    array_2_normalized = array_2
                else:
                    array_2_normalized = na.FunctionArray(inputs=array.inputs, outputs=array_2)
            else:
                if isinstance(array_2, na.AbstractFunctionArray):
                    array_normalized = na.FunctionArray(inputs=array_2.inputs, outputs=array)
                    array_2_normalized = array_2
                else:
                    raise TypeError("neither `array` nor `array_2` are instances of`named_arrays.FunctionArray`")

            try:
                outputs_expected = np.matmul(array_normalized.outputs, array_2_normalized.outputs)
                inputs_expected = array_normalized.inputs
            except Exception as e:
                with pytest.raises(type(e)):
                    np.matmul(array, array_2)
                return

            result = np.matmul(array, array_2)

            assert np.all(result.outputs == outputs_expected)
            assert np.all(result.inputs == inputs_expected)

    class TestArrayFunctions(
        named_arrays.tests.test_core.AbstractTestAbstractArray.TestArrayFunctions
    ):
        @pytest.mark.parametrize("array_2", _function_arrays_2())
        class TestAsArrayLikeFunctions(
            named_arrays.tests.test_core.AbstractTestAbstractArray.TestArrayFunctions.TestAsArrayLikeFunctions
        ):
            def test_asarray_like_functions(
                    self,
                    func: Callable,
                    array: None | float | u.Quantity | na.AbstractArray,
                    array_2: None | float | u.Quantity | na.AbstractArray,
            ):
                a = array
                like = array_2

                if a is None:
                    assert func(a, like=like) is None
                    return

                result = func(a, like=like)

                assert isinstance(result, na.FunctionArray)
                assert isinstance(result.inputs, na.AbstractArray)
                assert isinstance(result.outputs, na.AbstractArray)

                assert np.all(result.value == na.value(a))

                super().test_asarray_like_functions(
                    func=func,
                    array=array,
                    array_2=array_2,
                )

        @pytest.mark.skip
        class TestSingleArgumentFunctions(
            named_arrays.tests.test_core.AbstractTestAbstractArray.TestArrayFunctions.TestSingleArgumentFunctions,
        ):
            def test_single_argument_functions(
                self,
                func: Callable,
                array: na.AbstractFunctionArray,
            ):
                assert False

        @pytest.mark.xfail
        class TestArrayCreationLikeFunctions(
            named_arrays.tests.test_core.AbstractTestAbstractArray.TestArrayFunctions.TestArrayCreationLikeFunctions
        ):
            pass

        @pytest.mark.parametrize(
            argnames='where',
            argvalues=[
                np._NoValue,
                True,
                na.ScalarArray(True),
                na.FunctionArray(
                    inputs=na.Cartesian2dVectorLinearSpace(
                        start=0,
                        stop=1,
                        axis=na.Cartesian2dVectorArray('x', 'y'),
                        num=na.Cartesian2dVectorArray(_num_x, _num_y)
                    ),
                    outputs=(na.ScalarLinearSpace(-1, 1, 'x', _num_x) >= 0)
                            | (na.ScalarLinearSpace(-1, 1, 'y', _num_y) >= 0)
                )
            ]
        )
        class TestReductionFunctions(
            named_arrays.tests.test_core.AbstractTestAbstractArray.TestArrayFunctions.TestReductionFunctions
        ):

            def test_reduction_functions(
                    self,
                    func: Callable,
                    array: na.AbstractFunctionArray,
                    axis: None | str | Sequence[str],
                    dtype: None | type | np.dtype,
                    keepdims: bool,
                    where: bool | na.AbstractFunctionArray,
            ):

                array_broadcasted = na.broadcast_to(array, na.shape_broadcasted(array, where))

                where_normalized = where.outputs if isinstance(where, na.FunctionArray) else where

                kwargs = dict()
                kwargs_output = dict()
                if dtype is not np._NoValue:
                    kwargs["dtype"] = kwargs_output["dtype"] = dtype
                if func in [np.min, np.nanmin, np.max, np.nanmax]:
                    kwargs["initial"] = kwargs_output["initial"] = 0
                if where is not np._NoValue:
                    kwargs["where"] = where
                    kwargs_output["where"] = where_normalized

                try:
                    if isinstance(where, na.FunctionArray):
                        if np.any(where.inputs != array.inputs):
                            raise na.InputValueError
                    outputs_expected = func(
                        a=array_broadcasted.outputs,
                        axis=axis,
                        keepdims=keepdims,
                        **kwargs_output,
                    )
                    if keepdims:
                        inputs_expected = array_broadcasted.inputs
                    else:
                        inputs_expected = np.mean(
                            a=array_broadcasted.inputs,
                            axis=axis,
                            keepdims=keepdims,
                            where=where_normalized,
                        )
                except Exception as e:
                    with pytest.raises(type(e)):
                        func(array, axis=axis, keepdims=keepdims, **kwargs)
                    return

                result = func(array, axis=axis, keepdims=keepdims, **kwargs)

                out = 0 * result
                out.inputs = 0 * out.inputs
                result_out = func(array, axis=axis, out=out, keepdims=keepdims, **kwargs)

                assert np.all(result.inputs == inputs_expected)
                assert np.all(result.outputs == outputs_expected)
                assert np.all(result == result_out)
                assert result_out is out

        class TestPercentileLikeFunctions(
            named_arrays.tests.test_core.AbstractTestAbstractArray.TestArrayFunctions.TestPercentileLikeFunctions,
        ):

            @pytest.mark.parametrize(
                argnames='q',
                argvalues=[
                    .25,
                    25 * u.percent,
                    na.ScalarLinearSpace(.25, .75, axis='q', num=3, endpoint=True),
                    na.UncertainScalarArray(
                        nominal=25 * u.percent,
                        distribution=na.ScalarNormalRandomSample(
                            center=25 * u.percent,
                            width=1 * u.percent,
                            shape_random=dict(_distribution=_num_distribution)
                        )
                    )
                ]
            )
            def test_percentile_like_functions(
                    self,
                    func: Callable,
                    array: na.AbstractFunctionArray,
                    q: float | u.Quantity | na.AbstractArray,
                    axis: None | str | Sequence[str],
                    keepdims: bool,
            ):

                array_broadcasted = na.broadcast_to(array, array.shape)

                kwargs = dict()
                kwargs_output = dict()

                try:
                    outputs_expected = func(
                        a=array_broadcasted.outputs,
                        q=q,
                        axis=axis,
                        keepdims=keepdims,
                        **kwargs_output,
                    )
                    if keepdims:
                        inputs_expected = array_broadcasted.inputs
                    else:
                        inputs_expected = np.mean(
                            a=array_broadcasted.inputs,
                            axis=axis,
                            keepdims=keepdims,
                        )
                except Exception as e:
                    with pytest.raises(type(e)):
                        func(array, q=q, axis=axis, keepdims=keepdims, **kwargs)
                    return

                result = func(array, q=q, axis=axis, keepdims=keepdims, **kwargs)

                out = 0 * result
                out.inputs = 0 * out.inputs
                result_out = func(array, q=q, axis=axis, out=out, keepdims=keepdims, **kwargs)

                assert np.all(result.inputs == inputs_expected)
                assert np.all(result.outputs == outputs_expected)
                assert np.all(result == result_out)
                assert result_out is out

        @pytest.mark.skip
        class TestArgReductionFunctions(
            named_arrays.tests.test_core.AbstractTestAbstractArray.TestArrayFunctions.TestArgReductionFunctions
        ):

            def test_arg_reduction_functions(
                    self,
                    argfunc: Callable,
                    func: Callable,
                    array: na.AbstractFunctionArray,
                    axis: None | str,
            ):
                try:
                    outputs_expected = argfunc(array.outputs, axis=axis)
                except Exception as e:
                    with pytest.raises(type(e)):
                        argfunc(array, axis=axis)
                    return

                result = argfunc(array, axis=axis)

                assert np.all(result.outputs == outputs_expected)


        class TestFFTLikeFunctions(
            named_arrays.tests.test_core.AbstractTestAbstractArray.TestArrayFunctions.TestFFTLikeFunctions,
        ):

            @pytest.mark.skip
            def test_fft_like_functions(
                    self,
                    func: Callable,
                    array: na.AbstractFunctionArray,
                    axis: tuple[str, str],
            ):
                pass

        class TestFFTNLikeFunctions(
            named_arrays.tests.test_core.AbstractTestAbstractArray.TestArrayFunctions.TestFFTNLikeFunctions,
        ):

            @pytest.mark.skip
            def test_fftn_like_functions(
                    self,
                    func: Callable,
                    array: na.AbstractFunctionArray,
                    axes: dict[str, str],
                    s: None | dict[str, int],
            ):
                pass

        class TestEmathFunctions(
            named_arrays.tests.test_core.AbstractTestAbstractArray.TestArrayFunctions.TestEmathFunctions,
        ):

            @pytest.mark.skip
            def test_emath_functions(
                self,
                func: Callable,
                array: na.AbstractArray,
            ):
                pass

        @pytest.mark.xfail
        def test_sort(self, array: na.AbstractFunctionArray, axis: None | str | Sequence[str]):
            pass

        @pytest.mark.xfail
        def test_argsort(self, array: na.AbstractArray, axis: None | str | Sequence[str]):
            super().test_argsort(array=array, axis=axis)

        @pytest.mark.xfail
        def test_unravel_index(self, array: na.AbstractArray):
            super().test_unravel_index(array=array)

        @pytest.mark.skip
        def test_where(self, array: na.AbstractArray):
            super().test_where(array=array)

        @pytest.mark.skip
        def test_nan_to_num(self, array: na.AbstractFunctionArray, copy: bool):
            pass

        @pytest.mark.skip
        def test_convolve(self, array: na.AbstractFunctionArray, v: na.AbstractArray, mode: str):
            pass

        @pytest.mark.skip
        def test_diff_1st_order(
            self,
            array: na.AbstractArray,
            axis: str,
        ):
            pass    # pragma: nocover

    class TestNamedArrayFunctions(
        named_arrays.tests.test_core.AbstractTestAbstractArray.TestNamedArrayFunctions
    ):
        @pytest.mark.skip
        class TestInterp(
            named_arrays.tests.test_core.AbstractTestAbstractArray.TestNamedArrayFunctions.TestInterp,
        ):
            pass

        @pytest.mark.xfail
        class TestPltPlotLikeFunctions(
            named_arrays.tests.test_core.AbstractTestAbstractArray.TestNamedArrayFunctions.TestPltPlotLikeFunctions
        ):
            pass

        @pytest.mark.skip
        class TestPltScatter(
            named_arrays.tests.test_core.AbstractTestAbstractArray.TestNamedArrayFunctions.TestPltScatter
        ):
            pass

        class TestPltPcolormesh(
            named_arrays.tests.test_core.AbstractTestAbstractArray.TestNamedArrayFunctions.TestPltPcolormesh,
        ):
            @pytest.mark.parametrize("axis_rgb", [None, "rgb"])
            def test_pcolormesh(
                self,
                array: na.AbstractScalarArray,
                axis_rgb: None | str
            ):
                if not isinstance(array.inputs, na.AbstractVectorArray):
                    return

                components = list(array.inputs.components.keys())[:2]

                kwargs = dict(
                    C=array,
                    axis_rgb=axis_rgb,
                    components=components,
                )

                if isinstance(array.outputs, na.AbstractVectorArray):
                    with pytest.raises(TypeError):
                        na.plt.pcolormesh(**kwargs)
                    return
                elif isinstance(array.outputs, na.AbstractUncertainScalarArray):
                    with pytest.raises(TypeError):
                        na.plt.pcolormesh(**kwargs)
                    return

                if axis_rgb is not None:
                    with pytest.raises(ValueError):
                        na.plt.pcolormesh(**kwargs)
                    return

                result = na.plt.pcolormesh(**kwargs)
                assert isinstance(result, na.ScalarArray)

        @pytest.mark.xfail
        class TestJacobian(
            named_arrays.tests.test_core.AbstractTestAbstractArray.TestNamedArrayFunctions.TestJacobian,
        ):
            pass

        @pytest.mark.skip
        class TestOptimizeRoot(
            named_arrays.tests.test_core.AbstractTestAbstractArray.TestNamedArrayFunctions.TestOptimizeRoot,
        ):
            pass

        @pytest.mark.skip
        class TestOptimizeMinimum(
            named_arrays.tests.test_core.AbstractTestAbstractArray.TestNamedArrayFunctions.TestOptimizeMinimum,
        ):
            pass


@pytest.mark.parametrize("array", _function_arrays())
class TestFunctionArray(
    AbstractTestAbstractFunctionArray,
    named_arrays.tests.test_core.AbstractTestAbstractExplicitArray,
):

    @pytest.mark.parametrize(
        argnames="item",
        argvalues=[
            dict(y=0),
            dict(x=0, y=0),
            dict(y=slice(None)),
            dict(y=na.ScalarArrayRange(0, _num_y, axis='y')),
            dict(x=na.ScalarArrayRange(0, _num_x, axis='x'), y=na.ScalarArrayRange(0, _num_y, axis='y')),
            na.FunctionArray(
                inputs=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y),
                outputs=na.ScalarArray.ones(shape=dict(y=_num_y), dtype=bool),
            )
        ]
    )
    @pytest.mark.parametrize(
        argnames="value",
        argvalues=[
            0,
            na.FunctionArray(
                inputs=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y),
                outputs=na.ScalarUniformRandomSample(-5, 5, dict(y=_num_y)),
            )
        ]
    )
    def test__setitem__(
            self,
            array: na.AbstractArray,
            item: dict[str, int | slice | na.ScalarArray] | na.AbstractFunctionArray,
            value: float | u.Quantity | na.AbstractScalar | na.AbstractVectorArray,
    ):
        super().test__setitem__(array=array.explicit, item=item, value=value)


@pytest.mark.parametrize("type_array", [na.FunctionArray])
class TestFunctionArrayCreation(
    named_arrays.tests.test_core.AbstractTestAbstractExplicitArrayCreation,
):
    @pytest.mark.parametrize("like", [None] + _function_arrays())
    class TestFromScalarArray(
        named_arrays.tests.test_core.AbstractTestAbstractExplicitArrayCreation.TestFromScalarArray,
    ):
        pass



from typing import Callable
import numpy as np
import matplotlib
import astropy.units as u
import named_arrays as na
import named_arrays._scalars.scalar_named_array_functions

__all__ = [
    "ASARRAY_LIKE_FUNCTIONS",
]

ASARRAY_LIKE_FUNCTIONS = named_arrays._scalars.scalar_named_array_functions.ASARRAY_LIKE_FUNCTIONS
HANDLED_FUNCTIONS = dict()

def _implements(function: Callable):
    """Register a __named_array_function__ implementation for AbstractScalarArray objects."""
    def decorator(func):
        HANDLED_FUNCTIONS[function] = func
        return func
    return decorator


def asarray_like(
        func: Callable,
        a: None | float | u.Quantity | na.AbstractScalar | na.AbstractVectorArray | na.AbstractFunctionArray,
        dtype: None | type | np.dtype = None,
        order: None | str = None,
        *,
        like: None | float | u.Quantity | na.AbstractScalar | na.AbstractVectorArray | na.AbstractFunctionArray = None,
) -> None | na.AbstractFunctionArray:

    if isinstance(a, na.AbstractArray):
        if isinstance(a, na.AbstractFunctionArray):
            a_inputs = a.inputs
            a_outputs = a.outputs
        elif isinstance(a, na.AbstractVectorArray):
            a_inputs = a_outputs = a
        elif isinstance(a, na.AbstractScalar):
            a_inputs = a_outputs = a
        else:
            return NotImplemented
    else:
        a_inputs = a_outputs = a

    if isinstance(like, na.AbstractArray):
        if isinstance(like, na.AbstractFunctionArray):
            like_inputs = like.inputs
            like_outputs = like.outputs
            type_like = like.type_explicit
        elif isinstance(like, na.AbstractVectorArray):
            like_inputs = like_outputs = like
            type_like = na.FunctionArray
        elif isinstance(like, na.AbstractScalar):
            like_inputs = like_outputs = like
            type_like = na.FunctionArray
        else:
            return NotImplemented
    else:
        like_inputs = like_outputs = like
        type_like = na.FunctionArray

    return type_like(
        inputs=func(
            a=a_inputs,
            dtype=dtype,
            order=order,
            like=like_inputs,
        ),
        outputs=func(
            a=a_outputs,
            dtype=dtype,
            order=order,
            like=like_outputs,
        ),
    )


@_implements(na.unit)
def unit(
        a: na.AbstractFunctionArray,
        unit_dimensionless: None | float | u.UnitBase = None,
        squeeze: bool = True,
) -> None | u.UnitBase | na.AbstractArray:
    return na.unit(
        a=a.outputs,
        unit_dimensionless=unit_dimensionless,
        squeeze=squeeze,
    )


@_implements(na.unit_normalized)
def unit_normalized(
        a: na.AbstractFunctionArray,
        unit_dimensionless: float | u.UnitBase,
        squeeze: bool = True,
) -> u.UnitBase | na.AbstractArray:
    return na.unit_normalized(
        a.outputs,
        unit_dimensionless=unit_dimensionless,
        squeeze=squeeze,
    )


@_implements(na.plt.pcolormesh)
def pcolormesh(
    *XY: na.AbstractArray,
    C: na.AbstractFunctionArray,
    components: None | tuple[str, str] = None,
    axis_rgb: None | str = None,
    ax: None | matplotlib.axes.Axes | na.AbstractArray = None,
    cmap: None | str | matplotlib.colors.Colormap = None,
    norm: None | str | matplotlib.colors.Normalize = None,
    vmin: None | na.ArrayLike = None,
    vmax: None | na.ArrayLike = None,
    **kwargs,
) -> na.ScalarArray:

    if len(XY) != 0:    # pragma: nocover
        raise ValueError(
            "if `C` is an instance of `na.AbstractFunctionArray`, "
            "`XY` must not be specified."
        )

    return na.plt.pcolormesh(
        C.inputs,
        C=C.outputs,
        components=components,
        axis_rgb=axis_rgb,
        ax=ax,
        cmap=cmap,
        norm=norm,
        vmin=vmin,
        vmax=vmax,
        **kwargs,
    )


@_implements(na.ndfilters.mean_filter)
def mean_filter(
    array: na.AbstractFunctionArray,
    size: dict[str, int],
    where: bool | na.AbstractFunctionArray,
) -> na.FunctionArray:

    if isinstance(array, na.AbstractFunctionArray):
        pass
    else:
        return NotImplemented   # pragma: nocover

    if isinstance(where, bool):
        where = na.FunctionArray(None, where)
    elif isinstance(where, na.AbstractFunctionArray):
        if np.all(where.inputs != array.inputs):    # pragma: nocover
            raise ValueError(
                f"if `where` is an instance of `na.AbstractFunctionArray`, "
                f"its inputs must match `array`."
            )
    else:
        return NotImplemented   # pragma: nocover

    return array.type_explicit(
        inputs=array.inputs.copy(),
        outputs=na.ndfilters.mean_filter(
            array=array.outputs,
            size=size,
            where=where.outputs,
        )
    )



from __future__ import annotations
from typing import TypeVar
import named_arrays as na

__all__ = [
    "mean_filter",
]

ArrayT = TypeVar("ArrayT", bound="na.AbstractArray")
WhereT = TypeVar("WhereT", bound="bool | na.AbstractArray")


def mean_filter(
    array: ArrayT,
    size: dict[str, int],
    where: WhereT = True,
) -> ArrayT | WhereT:
    """
    A thin wrapper around :func:`ndfilters.mean_filter` for named arrays.

    Parameters
    ----------
    array
        The input array to be filtered.
    size
        The shape of the kernel over which the mean will be calculated.
    where
        A boolean mask used to select which elements of the input array are to
        be filtered.

    Examples
    --------

    Filter a sample image.

    .. jupyter-execute::
        :stderr:

        import matplotlib.pyplot as plt
        import scipy.datasets
        import named_arrays as na

        img = na.ScalarArray(scipy.datasets.ascent(), axes=("x", "y"))

        img_filtered = na.ndfilters.mean_filter(img, size=dict(x=21, y=21))

        fig, axs = plt.subplots(
            ncols=2,
            sharex=True,
            sharey=True,
            constrained_layout=True,
        )
        axs[0].set_title("original image");
        na.plt.imshow(
            X=img,
            axis_x="x",
            axis_y="y",
            ax=axs[0],
            cmap="gray",
        );
        axs[1].set_title("mean filtered image");
        na.plt.imshow(
            X=img_filtered,
            axis_x="x",
            axis_y="y",
            ax=axs[1],
            cmap="gray",
        );
    """
    return na._named_array_function(
        func=mean_filter,
        array=array,
        size=size,
        where=where,
    )



from __future__ import annotations
from typing import Type, TypeVar
from typing_extensions import Self
import dataclasses
import named_arrays as na

__all__ = [
    'AbstractTemporalSpectralPositionalVectorArray',
    'TemporalSpectralPositionalVectorArray',
    'AbstractImplicitTemporalSpectralPositionalVectorArray',
    'AbstractParameterizedTemporalSpectralPositionalVectorArray',
    'AbstractTemporalSpectralPositionalVectorSpace',
    'TemporalSpectralPositionalVectorLinearSpace',
    'ExplicitTemporalWcsSpectralPositionalVectorArray',
]

TimeT = TypeVar("TimeT", bound=na.ArrayLike)
PositionT = TypeVar("PositionT", bound=na.ArrayLike)
WavelengthT = TypeVar("WavelengthT", bound=na.ScalarLike)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractTemporalSpectralPositionalVectorArray(
    na.AbstractPositionalVectorArray,
    na.AbstractSpectralVectorArray,
    na.AbstractTemporalVectorArray,
):

    @property
    def type_abstract(self) -> Type[na.AbstractArray]:
        return AbstractTemporalSpectralPositionalVectorArray

    @property
    def type_explicit(self) -> Type[na.AbstractExplicitArray]:
        return TemporalSpectralPositionalVectorArray

    @property
    def type_matrix(self) -> Type[na.TemporalSpectralPositionalMatrixArray]:
        return na.TemporalSpectralPositionalMatrixArray


@dataclasses.dataclass(eq=False, repr=False)
class TemporalSpectralPositionalVectorArray(
    AbstractTemporalSpectralPositionalVectorArray,
    na.PositionalVectorArray[PositionT],
    na.SpectralVectorArray[WavelengthT],
    na.TemporalVectorArray
):

    @classmethod
    def from_scalar(
            cls: Type[Self],
            scalar: na.AbstractScalar,
            like: None | na.AbstractExplicitVectorArray = None,
    ) -> TemporalSpectralPositionalVectorArray:
        return cls(time=scalar, wavelength=scalar, position=scalar)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImplicitTemporalSpectralPositionalVectorArray(
    AbstractTemporalSpectralPositionalVectorArray,
    na.AbstractImplicitPositionalVectorArray,
    na.AbstractImplicitSpectralVectorArray,
    na.AbstractImplicitTemporalVectorArray,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractParameterizedTemporalSpectralPositionalVectorArray(
    AbstractImplicitTemporalSpectralPositionalVectorArray,
    na.AbstractParameterizedPositionalVectorArray,
    na.AbstractParameterizedSpectralVectorArray,
    na.AbstractParameterizedTemporalVectorArray,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractTemporalSpectralPositionalVectorSpace(
    AbstractParameterizedTemporalSpectralPositionalVectorArray,
    na.AbstractPositionalVectorSpace,
    na.AbstractSpectralVectorSpace,
    na.AbstractTemporalVectorSpace
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class TemporalSpectralPositionalVectorLinearSpace(
    AbstractTemporalSpectralPositionalVectorSpace,
    na.AbstractVectorLinearSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class ExplicitTemporalWcsSpectralPositionalVectorArray(
    AbstractImplicitTemporalSpectralPositionalVectorArray,
    na.AbstractWcsVector,
):
    time: na.AbstractScalar = dataclasses.MISSING
    crval: AbstractTemporalSpectralPositionalVectorArray = dataclasses.MISSING
    crpix: na.AbstractCartesianNdVectorArray = dataclasses.MISSING
    cdelt: AbstractTemporalSpectralPositionalVectorArray = dataclasses.MISSING
    pc: na.AbstractTemporalSpectralPositionalMatrixArray = dataclasses.MISSING
    shape_wcs: dict[str, int] = dataclasses.MISSING

    @property
    def _components_explicit(self) -> dict[str, na.ArrayLike]:
        return dict(time=self.time)



from __future__ import annotations
from typing import Type

import dataclasses
import named_arrays as na

__all__ = [
    'AbstractTemporalSpectralDirectionalMatrixArray',
    'TemporalSpectralDirectionalMatrixArray',
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractTemporalSpectralDirectionalMatrixArray(
    na.AbstractTemporalSpectralDirectionalVectorArray,
):

    @property
    def type_abstract(self) -> Type[AbstractTemporalSpectralDirectionalMatrixArray]:
        return AbstractTemporalSpectralDirectionalMatrixArray

    @property
    def type_explicit(self) -> Type[TemporalSpectralDirectionalMatrixArray]:
        return TemporalSpectralDirectionalMatrixArray

    @property
    def type_vector(self) -> Type[na.TemporalSpectralDirectionalVectorArray]:
        return na.TemporalSpectralDirectionalVectorArray

    @property
    def determinant(self) -> na.ScalarLike:
        return NotImplementedError


@dataclasses.dataclass(eq=False, repr=False)
class TemporalSpectralDirectionalMatrixArray(
    na.TemporalSpectralDirectionalVectorArray,
    AbstractTemporalSpectralDirectionalMatrixArray,
    na.AbstractExplicitMatrixArray,
):
    pass



from __future__ import annotations
from typing import Type, Sequence, Callable
import pytest
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import astropy.units as u
import astropy.visualization
import astropy.units.quantity_helper.helpers as quantity_helpers
import named_arrays as na
from ... import tests

__all__ = [
    'AbstractTestAbstractScalar',
    'AbstractTestAbstractScalarArray',
    'TestScalarArray',
    'TestScalarArrayCreation',
    'AbstractTestAbstractParameterizedScalarArray',
]

_num_x = tests.test_core.num_x
_num_y = tests.test_core.num_y
_num_distribution = tests.test_core.num_distribution


def _scalar_arrays():
    arrays_numeric = [
        na.ScalarArray(4),
        na.ScalarArray(5.),
        na.ScalarArray(10 * (np.random.random((_num_y, )) - 0.5), axes=('y', )),
        na.ScalarArray(10 * (np.random.random((_num_x, _num_y)) - 0.5), axes=('x', 'y')),
    ]
    units = [1, u.mm]
    arrays_numeric = [na.ScalarArray(array.ndarray * unit, array.axes) for array in arrays_numeric for unit in units]
    arrays_bool = [
        na.ScalarArray(np.random.choice([True, False], size=_num_y), axes=('y', )),
        na.ScalarArray(np.random.choice([True, False], size=(_num_x, _num_y)), axes=('x', 'y'))
    ]
    arrays_bool[0][dict(y=0)] = True
    arrays_bool[1][dict(x=0, y=0)] = True
    return arrays_numeric + arrays_bool


def _scalar_arrays_2():
    arrays_numeric = [
        6,
        na.ScalarArray(8),
        na.ScalarArray(10 * (np.random.random((_num_y,)) - 0.5), axes = ('y', )),
        na.ScalarArray(10 * (np.random.random((_num_y, _num_x)) - 0.5), axes=('y', 'x')),
    ]
    units = [1, u.m]
    arrays_numeric = [array * unit for array in arrays_numeric for unit in units]
    arrays_bool = [
        na.ScalarArray(np.random.choice([True, False], size=_num_y), axes=('y', )),
        na.ScalarArray(np.random.choice([True, False], size=(_num_y, _num_x)), axes=('y', 'x'))
    ]
    return [None] + arrays_numeric + arrays_bool


@pytest.mark.parametrize('value', _scalar_arrays_2())
def test_as_named_array(value: bool | int | float | complex | str | u.Quantity | na.AbstractArray):
    result = na.as_named_array(value)
    assert isinstance(result, na.AbstractArray)


class AbstractTestAbstractScalar(
    tests.test_core.AbstractTestAbstractArray,
):

    def test_dtype(self, array: na.AbstractScalar):
        assert array.dtype == array.explicit.dtype
        assert isinstance(array.dtype, np.dtype)

    def test_unit(self, array: na.AbstractScalar):
        assert array.unit == array.explicit.unit
        unit = array.unit
        if unit is not None:
            assert isinstance(unit, u.UnitBase)

    def test_unit_normalized(self, array: na.AbstractScalar):
        assert array.unit_normalized == array.explicit.unit_normalized
        if array.unit is None:
            assert array.unit_normalized == u.dimensionless_unscaled
        else:
            assert array.unit_normalized == array.unit

    @pytest.mark.parametrize('dtype', [int, float])
    def test_astype(self, array: na.AbstractScalar, dtype: Type):
        super().test_astype(array=array, dtype=dtype)
        array_new = array.astype(dtype)
        assert array_new.dtype == dtype

    @pytest.mark.parametrize('unit', [u.m, u.s])
    def test_to(self, array: na.AbstractScalar, unit: None | u.UnitBase):
        super().test_to(array=array, unit=unit)
        if isinstance(array.unit, u.UnitBase) and array.unit.is_equivalent(unit):
            array_new = array.to(unit)
            assert array_new.unit == unit
        else:
            with pytest.raises(u.UnitConversionError):
                array.to(unit)

    def test_length(self, array: na.AbstractScalar):
        super().test_length(array=array)
        if not np.issubdtype(array.dtype, np.number):
            with pytest.raises(ValueError):
                array.length
            return

        length = array.length
        assert isinstance(length, (int, float, np.ndarray, na.AbstractScalar))
        assert np.all(length >= 0)

    @pytest.mark.parametrize(
        argnames="axis",
        argvalues=[
            None,
            "x",
            "y",
            ("y", ),
            ("x", "y"),
        ]
    )
    def test_volume_cell(
            self,
            array: na.AbstractArray,
            axis: None | str | Sequence[str],
    ):
        axis_ = na.axis_normalized(array, axis)

        if not set(axis_).issubset(array.axes):
            with pytest.raises(ValueError):
                array.volume_cell(axis)
            return

        if len(axis_) != 1:
            with pytest.raises(ValueError):
                array.volume_cell(axis)
            return

        result = array.volume_cell(axis)
        shape_result = na.shape(result)

        for ax in array.shape:
            if ax in shape_result:
                if ax in axis_:
                    assert shape_result[ax] == array.shape[ax] - 1
                else:
                    assert shape_result[ax] == array.shape[ax]

        assert isinstance(na.as_named_array(result), na.AbstractScalar)

    def test__bool__(self, array: na.AbstractScalarArray):
        if array.shape or array.unit is not None:
            with pytest.raises(
                expected_exception=ValueError,
                match=r"(Quantity truthiness is ambiguous, .*)"
                      r"|(The truth value of an array with more than one element is ambiguous. .*)"
            ):
                bool(array)
            return

        result = bool(array)
        assert isinstance(result, bool)

    class TestMatmul(
        tests.test_core.AbstractTestAbstractArray.TestMatmul,
    ):

        def test_matmul(
                self,
                array: None | bool | int | float | complex | str | na.AbstractArray,
                array_2: None | bool | int | float | complex | str | na.AbstractArray,
        ):
            result = np.matmul(array, array_2)

            if array is None or array_2 is None:
                assert result is None
                return

            result_expected = np.multiply(array, array_2)

            out = 0 * result
            result_out = np.matmul(array, array_2, out=out)

            assert np.all(result == result_expected)
            assert np.all(result == result_out)
            assert result_out is out

    class TestNamedArrayFunctions(
        tests.test_core.AbstractTestAbstractArray.TestNamedArrayFunctions,
    ):
        @pytest.mark.parametrize(
            argnames="bins,min,max",
            argvalues=[
                (
                    dict(hx=11, hy=12),
                    -100,
                    100,
                ),
                (
                    na.Cartesian2dVectorArray(
                        x=na.linspace(-100, 100, axis="hx", num=11),
                        y=na.linspace(-100, 100, axis="hy", num=11),
                    ),
                    None,
                    None,
                )
            ]
        )
        @pytest.mark.parametrize(
            argnames="axis",
            argvalues=[
                None,
                "y",
            ]
        )
        @pytest.mark.parametrize(
            argnames="weights",
            argvalues=[
                None,
            ]
        )
        class TestHistogram2d:
            def test_histogram2d(
                self,
                array: na.AbstractScalar,
                bins: dict[str, int] | na.AbstractCartesian2dVectorArray,
                axis: None | str | Sequence[str],
                min: None | na.AbstractScalarArray | na.AbstractCartesian2dVectorArray,
                max: None | na.AbstractScalarArray | na.AbstractCartesian2dVectorArray,
                weights: None | na.AbstractScalarArray,
            ):
                if not array.shape:
                    return

                if isinstance(bins, na.AbstractCartesian2dVectorArray):
                    bins = bins * array.unit_normalized
                if min is not None:
                    min = min * array.unit_normalized
                if max is not None:
                    max = max * array.unit_normalized

                result = na.histogram2d(
                    x=array,
                    y=array,
                    bins=bins,
                    axis=axis,
                    min=min,
                    max=max,
                    weights=weights,
                )

                assert np.all(result.outputs >= 0)
                assert result.outputs.sum() > 0


class AbstractTestAbstractScalarArray(
    AbstractTestAbstractScalar,
):

    def test_ndarray(self, array: na.AbstractScalarArray):
        assert isinstance(array.ndarray, (int, float, complex, str, np.ndarray))

    def test_axes(self, array: na.AbstractScalarArray):
        super().test_axes(array)
        assert len(array.axes) == np.ndim(array.ndarray)

    def test_size(self, array: na.AbstractScalarArray):
        super().test_size(array)
        assert array.size == np.size(array.ndarray)

    @pytest.mark.parametrize('index', [1, ~0])
    def test_change_axis_index(self, array: na.ScalarArray, index: int):
        axis = 'x'
        if axis in array.axes:
            result = array.change_axis_index(axis, index)
            assert result.axes.index(axis) == (index % array.ndim)
        else:
            with pytest.raises(KeyError):
                array.change_axis_index(axis, index)

    @pytest.mark.parametrize(
        argnames='item',
        argvalues=[
            dict(y=0),
            dict(y=slice(0,1)),
            dict(y=na.ScalarArray(np.array([0, 1]), axes=('y', ))),
            na.ScalarLinearSpace(0, 1, axis='y', num=_num_y) > 0.5,
        ]
    )
    def test__getitem__(
            self,
            array: na.AbstractScalarArray,
            item: dict[str, int | slice | na.AbstractArray] | na.AbstractArray
    ):
        super().test__getitem__(array=array, item=item)

        if array.shape:
            result = array[item]
            assert isinstance(result, na.AbstractArray)

            if isinstance(item, dict):
                item_expected = [Ellipsis]
                for axis in item:
                    if isinstance(item[axis], na.AbstractArray):
                        item_expected.append(item[axis].ndarray)
                    else:
                        item_expected.append(item[axis])
                item_expected = tuple(item_expected)
            else:
                item_expected = (Ellipsis, item.ndarray)

            result_expected = array.ndarray[item_expected]
            if 'y' in result.axes:
                result = result.change_axis_index('y', ~0)
            assert np.all(result.ndarray == result_expected)

        else:
            with pytest.raises(ValueError):
                array[item]

    @pytest.mark.parametrize(
        argnames='item',
        argvalues=[
            dict(
                y=na.UncertainScalarArray(
                    nominal=na.ScalarArray(np.array([0, 1]), axes=('y',)),
                    distribution=na.ScalarArray(
                        ndarray=np.array([[0, ], [1, ]]),
                        axes=('y', na.UncertainScalarArray.axis_distribution),
                    )
                ),
                _distribution=na.UncertainScalarArray(
                    nominal=None,
                    distribution=na.ScalarArray(
                        ndarray=np.array([[0], [0]]),
                        axes=('y', na.UncertainScalarArray.axis_distribution),
                    )
                )
            ),
            na.UniformUncertainScalarArray(
                nominal=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y),
                width=.1,
                num_distribution=_num_distribution,
            ) > 0.5,
        ]
    )
    def test__getitem__uncertain(
            self,
            array: na.AbstractScalarArray,
            item: dict[str, na.UncertainScalarArray] | na.UncertainScalarArray
    ):

        if isinstance(item, dict):

            if not (set(item) - set((na.UncertainScalarArray.axis_distribution, ))).issubset(array.shape):
                with pytest.raises(
                    expected_exception=ValueError,
                    match="the axes in item, .*, must be a subset of the axes in the array, .*"
                ):
                    array[item]
                return

            num_distribution = item[na.UncertainScalarArray.axis_distribution].distribution.max().ndarray + 1
            shape_distribution = na.broadcast_shapes(
                array.shape,
                {na.UncertainScalarArray.axis_distribution: num_distribution}
            )

            result_expected = na.UncertainScalarArray(
                nominal=array[{ax: item[ax].nominal for ax in item}],
                distribution=array.broadcast_to(shape_distribution)[{ax: item[ax].distribution for ax in item}],
            )
        else:
            if not set(item.shape).issubset(array.shape):
                with pytest.raises(
                    expected_exception=ValueError,
                    match="the axes in item, .*, must be a subset of the axes in array, .*"
                ):
                    array[item]
                return

            result_expected = array[item.nominal & np.all(item.distribution, axis=item.axis_distribution)]

        result = array[item]

        assert np.all(result == result_expected)

    def test__mul__(self, array: na.AbstractScalarArray):
        unit = u.mm
        result = array * unit
        result_ndarray = array.ndarray * unit
        assert np.all(result.ndarray == result_ndarray)

    def test__lshift__(self, array: na.AbstractScalarArray):
        unit = u.mm
        result = array << unit
        result_ndarray = array.ndarray << unit
        assert np.all(result.ndarray == result_ndarray)

    def test__truediv__(self, array: na.AbstractScalarArray):
        unit = u.mm
        result = array / unit
        result_ndarray = array.ndarray / unit
        assert np.all(result.ndarray == result_ndarray)

    class TestUfuncUnary(
        AbstractTestAbstractScalar.TestUfuncUnary,
    ):

        def test_ufunc_unary(
                self,
                ufunc: np.ufunc,
                array: na.AbstractScalarArray,
        ):
            super().test_ufunc_unary(ufunc, array)

            kwargs = dict()
            kwargs_ndarray = dict()

            unit = array.unit_normalized
            if ufunc in [np.log, np.log2, np.log10, np.sqrt]:
                kwargs["where"] = array > 0
            elif ufunc in [np.log1p]:
                kwargs["where"] = array >= (-1 * unit)
            elif ufunc in [np.arcsin, np.arccos, np.arctanh]:
                kwargs["where"] = ((-1 * unit) < array) & (array < (1 * unit))
            elif ufunc in [np.arccosh]:
                kwargs["where"] = array >= (1 * unit)
            elif ufunc in [np.reciprocal]:
                kwargs["where"] = array != 0

            if "where" in kwargs:
                kwargs_ndarray["where"] = kwargs["where"].ndarray

            try:
                ufunc(array.ndarray, **kwargs_ndarray)
            except (ValueError, TypeError) as e:
                with pytest.raises(type(e)):
                    ufunc(array, **kwargs)
                return

            result = ufunc(array, **kwargs)
            result_ndarray = ufunc(array.ndarray, **kwargs_ndarray)

            if ufunc.nout == 1:
                out = 0 * np.nan_to_num(result)
            else:
                out = tuple(0 * np.nan_to_num(r) for r in result)

            result_out = ufunc(array, out=out, **kwargs)

            if ufunc.nout == 1:
                out = (out, )
                result_ndarray = (result_ndarray, )
                result = (result, )
                result_out = (result_out, )

            for i in range(ufunc.nout):
                assert np.all(result[i].ndarray == result_ndarray[i], **kwargs_ndarray)
                assert np.all(result[i] == result_out[i], **kwargs)
                assert result_out[i] is out[i]

    @pytest.mark.parametrize('array_2', _scalar_arrays_2())
    class TestUfuncBinary(
        AbstractTestAbstractScalar.TestUfuncBinary,
    ):

        def test_ufunc_binary(
                self,
                ufunc: np.ufunc,
                array: None | bool | int | float | complex | str | na.AbstractScalarArray,
                array_2: None | bool | int | float | complex | str | na.AbstractScalarArray,
        ):
            super().test_ufunc_binary(ufunc, array, array_2)

            if array is None or array_2 is None:
                assert ufunc(array, array_2) is None
                return

            array_ndarray = array.ndarray if isinstance(array, na.AbstractArray) else array
            array_2_ndarray = array_2.ndarray if isinstance(array_2, na.AbstractArray) else array_2
            array_2_ndarray = np.transpose(array_2_ndarray)

            kwargs = dict()
            kwargs_ndarray = dict()

            unit_2 = na.unit_normalized(array_2)
            if ufunc in [np.power, np.float_power]:
                kwargs["where"] = (array_2 >= (1 * unit_2)) & (array >= 0)
            elif ufunc in [np.divide, np.floor_divide, np.remainder, np.fmod, np.divmod]:
                kwargs["where"] = array_2 != 0

            shape = na.shape_broadcasted(array, array_2)
            shape = {ax: shape[ax] for ax in sorted(shape)}
            if "where" in kwargs:
                kwargs["where"] = na.broadcast_to(kwargs["where"], shape)
                kwargs_ndarray["where"] = kwargs["where"].ndarray

            try:
                ufunc(array_ndarray, array_2_ndarray, **kwargs_ndarray)
            except (ValueError, TypeError) as e:
                with pytest.raises(type(e)):
                    ufunc(array, array_2, **kwargs)
                return

            result = ufunc(array, array_2, **kwargs)
            result_ndarray = ufunc(array_ndarray, array_2_ndarray, **kwargs_ndarray)

            if ufunc.nout == 1:
                out = 0 * np.nan_to_num(result)
            else:
                out = tuple(0 * np.nan_to_num(r) for r in result)

            result_out = ufunc(array, array_2, out=out, **kwargs)

            if ufunc.nout == 1:
                out = (out,)
                result_ndarray = (result_ndarray,)
                result = (result,)
                result_out = (result_out,)

            for i in range(ufunc.nout):
                assert np.all(result[i].broadcast_to(shape).ndarray == result_ndarray[i], **kwargs_ndarray)
                assert np.all(result[i] == result_out[i], **kwargs)
                assert result_out[i] is out[i]

    @pytest.mark.parametrize('array_2', _scalar_arrays_2())
    class TestMatmul(
        AbstractTestAbstractScalar.TestMatmul,
    ):
        pass

    class TestArrayFunctions(
        AbstractTestAbstractScalar.TestArrayFunctions,
    ):

        @pytest.mark.parametrize("array_2", _scalar_arrays_2())
        class TestAsArrayLikeFunctions(
            AbstractTestAbstractScalar.TestArrayFunctions.TestAsArrayLikeFunctions,
        ):

            def test_asarray_like_functions(
                    self,
                    func: Callable,
                    array: None | float | u.Quantity | na.AbstractArray,
                    array_2: None | float | u.Quantity | na.AbstractArray,
            ):
                a = array
                like = array_2

                if a is None:
                    assert func(a, like=like) is None
                    return

                result = func(a, like=like)

                assert isinstance(result, na.ScalarArray)
                assert isinstance(result.ndarray, np.ndarray)

                assert np.all(result.value == na.value(a))

                super().test_asarray_like_functions(
                    func=func,
                    array=array,
                    array_2=array_2,
                )

        class TestSingleArgumentFunctions(
            AbstractTestAbstractScalar.TestArrayFunctions.TestSingleArgumentFunctions,
        ):
            def test_single_argument_functions(
                self,
                func: Callable,
                array: na.AbstractScalarArray,
            ):
                result = func(array)
                assert np.all(result.ndarray == func(array.ndarray))

        @pytest.mark.parametrize(
            argnames='where',
            argvalues=[
                np._NoValue,
                True,
                na.ScalarArray(True),
                (na.ScalarLinearSpace(-1, 1, 'x', _num_x) >= 0) | (na.ScalarLinearSpace(-1, 1, 'y', _num_y) >= 0),
            ]
        )
        class TestReductionFunctions(
            AbstractTestAbstractScalar.TestArrayFunctions.TestReductionFunctions
        ):

            def test_reduction_functions(
                    self,
                    func: Callable,
                    array: na.AbstractScalarArray,
                    axis: None | str | Sequence[str],
                    dtype: None | type | np.dtype,
                    keepdims: bool,
                    where: bool | na.AbstractArray,
            ):
                super().test_reduction_functions(
                    func=func,
                    array=array,
                    axis=axis,
                    dtype=dtype,
                    keepdims=keepdims,
                    where=where,
                )

                kwargs = dict(
                    axis=axis,
                    keepdims=keepdims,
                    where=where,
                )

                axis_normalized = na.axis_normalized(array, axis=axis)

                if axis is not None:
                    if not set(axis_normalized).issubset(array.axes):
                        with pytest.raises(
                                ValueError, match=r"the `axis` argument must be `None` or a subset of"):
                            func(array, axis=axis)
                        return

                kwargs_ndarray = dict(
                    axis=tuple(array.axes.index(ax) for ax in axis_normalized),
                    keepdims=keepdims,
                    where=where.ndarray if isinstance(where, na.AbstractArray) else where,
                )

                if dtype is not np._NoValue:
                    kwargs["dtype"] = kwargs_ndarray["dtype"] = dtype

                if func in [np.min, np.nanmin, np.max, np.nanmax]:
                    kwargs["initial"] = kwargs_ndarray["initial"] = 0

                try:
                    result_ndarray = func(array.ndarray, **kwargs_ndarray)
                except (ValueError, TypeError, u.UnitsError) as e:
                    with pytest.raises(type(e)):
                        func(array, **kwargs)
                    return

                result = func(array, **kwargs)

                out = 0 * result
                result_out = func(array, out=out, **kwargs)

                assert np.all(result.ndarray == result_ndarray)
                assert np.allclose(result, result_out)
                assert result_out is out

        class TestPercentileLikeFunctions(
            AbstractTestAbstractScalar.TestArrayFunctions.TestPercentileLikeFunctions
        ):

            @pytest.mark.parametrize(
                argnames='q',
                argvalues=[
                    .25,
                    25 * u.percent,
                    na.ScalarLinearSpace(.25, .75, axis='q', num=3, endpoint=True),
                ]
            )
            def test_percentile_like_functions(
                    self,
                    func: Callable,
                    array: na.AbstractScalarArray,
                    q: float | u.Quantity | na.AbstractArray,
                    axis: None | str | Sequence[str],
                    keepdims: bool,
            ):
                super().test_percentile_like_functions(
                    func=func,
                    array=array.explicit,
                    q=q,
                    axis=axis,
                    keepdims=keepdims,
                )

                kwargs = dict(
                    q=q,
                    axis=axis,
                    keepdims=keepdims,
                )

                q_normalized = q if isinstance(q, na.AbstractArray) else na.ScalarArray(q)
                axis_normalized = na.axis_normalized(array, axis)

                if axis is not None:
                    if not set(axis_normalized).issubset(array.axes):
                        with pytest.raises(
                                ValueError, match=r"the `axis` argument must be `None` or a subset of"):
                            func(array, **kwargs)
                        return

                kwargs_ndarray = dict(
                    q=q_normalized.ndarray,
                    axis=tuple(array.axes.index(ax) for ax in axis_normalized),
                    keepdims=keepdims,
                )

                try:
                    result_ndarray = func(array.ndarray, **kwargs_ndarray)
                except (ValueError, TypeError) as e:
                    with pytest.raises(type(e)):
                        func(array, **kwargs)
                    return

                result = func(array, **kwargs)

                out = 0 * result
                result_out = func(array, out=out, **kwargs)

                assert np.all(result.ndarray == result_ndarray)
                assert np.allclose(result, result_out)
                assert result_out is out

            @pytest.mark.parametrize(
                argnames="q",
                argvalues=[
                    na.UncertainScalarArray(
                        nominal=25 * u.percent,
                        distribution=na.ScalarNormalRandomSample(
                            center=25 * u.percent,
                            width=1 * u.percent,
                            shape_random=dict(_distribution=_num_distribution)
                        )
                    ),
                    na.NormalUncertainScalarArray(
                        nominal=na.ScalarLinearSpace(25 * u.percent, 75 * u.percent, axis='y', num=_num_y),
                        width=1 * u.percent,
                        num_distribution=_num_distribution,
                    )
                ]
            )
            def test_percentile_like_functions_q_uncertain(
                    self,
                    func: Callable,
                    array: na.AbstractScalarArray,
                    q: float | u.Quantity | na.AbstractArray,
                    axis: None | str | Sequence[str],
                    keepdims: bool,
            ):
                from named_arrays._scalars.uncertainties.tests.test_uncertainties import (
                    AbstractTestAbstractUncertainScalarArray
                )
                other = AbstractTestAbstractUncertainScalarArray.TestArrayFunctions.TestPercentileLikeFunctions()
                other.test_percentile_like_functions(
                    func=func,
                    array=array,
                    q=q,
                    axis=axis,
                    keepdims=keepdims,
                )

        class TestFFTLikeFunctions(
            AbstractTestAbstractScalar.TestArrayFunctions.TestFFTLikeFunctions
        ):

            def test_fft_like_functions(
                    self,
                    func: Callable,
                    array: na.AbstractScalarArray,
                    axis: tuple[str, str],
            ):
                super().test_fft_like_functions(
                    func=func,
                    array=array,
                    axis=axis,
                )

                if axis[0] not in array.axes:
                    with pytest.raises(ValueError):
                        func(a=array, axis=axis)
                    return

                result = func(
                    a=array,
                    axis=axis,
                )
                result_expected = func(
                    a=array.ndarray,
                    axis=array.axes.index(axis[0])
                )

                assert isinstance(result, na.AbstractArray)
                assert axis[1] in result.axes
                assert not axis[0] in result.axes

                assert np.all(result.ndarray == result_expected)

        class TestFFTNLikeFunctions(
            AbstractTestAbstractScalar.TestArrayFunctions.TestFFTNLikeFunctions
        ):

            def test_fftn_like_functions(
                    self,
                    func: Callable,
                    array: na.AbstractScalarArray,
                    axes: dict[str, str],
                    s: None | dict[str, int],
            ):
                super().test_fftn_like_functions(
                    func=func,
                    array=array,
                    axes=axes,
                    s=s,
                )

                shape = array.shape

                if not all(ax in shape for ax in axes):
                    with pytest.raises(ValueError):
                        func(a=array, axes=axes, s=s)
                    return

                if func in [np.fft.rfft2, np.fft.irfft2, np.fft.rfftn, np.fft.irfftn]:
                    if not shape:
                        with pytest.raises(IndexError):
                            func(a=array, axes=axes, s=s)
                        return

                if s is not None and axes.keys() != s.keys():
                    with pytest.raises(ValueError):
                        func(a=array, axes=axes, s=s)
                    return

                if s is None:
                    s_normalized = {ax: shape[ax] for ax in axes}
                else:
                    s_normalized = {ax: s[ax] for ax in axes}

                result = func(
                    a=array,
                    axes=axes,
                    s=s,
                )
                result_expected = func(
                    a=array.ndarray,
                    s=s_normalized.values(),
                    axes=[array.axes.index(ax) for ax in axes],
                )

                assert isinstance(result, na.AbstractArray)
                assert all(axes[ax] in result.axes for ax in axes)
                assert all(ax not in result.axes for ax in axes)

                assert np.all(result.ndarray == result_expected)

        class TestEmathFunctions(
            AbstractTestAbstractScalar.TestArrayFunctions.TestEmathFunctions,
        ):
            def test_emath_functions(
                self,
                func: Callable,
                array: na.AbstractScalarArray,
            ):
                result = func(array)
                result_expected = func(array.ndarray)
                assert np.all(result.ndarray == result_expected)
                assert result.axes == array.axes

        @pytest.mark.parametrize('axis', [None, 'x', 'y', ('x', 'y')])
        def test_sort(self, array: na.AbstractScalarArray, axis: None | str | Sequence[str]):

            super().test_sort(array=array, axis=axis)

            axis_normalized = na.axis_normalized(array, axis)

            if axis is not None:
                if not set(axis_normalized).issubset(array.axes):
                    with pytest.raises(
                            expected_exception=ValueError,
                            match="`axis`, .* is not a subset of `a.axes`, .*"
                    ):
                        np.sort(array, axis=axis)
                    return
            else:
                if not array.shape:
                    result = np.sort(a=array, axis=axis)
                    assert np.all(result == array)
                return

            axis_flattened = na.flatten_axes(axis_normalized)
            array_normalized = array.combine_axes(
                axes=axis_normalized,
                axis_new=axis_flattened,
            )

            result = np.sort(a=array, axis=axis)
            result_ndarray = np.sort(
                a=array_normalized.ndarray,
                axis=array_normalized.axes.index(axis_flattened) if axis is not None else axis,
            )

            assert np.all(result.ndarray == result_ndarray)

        def test_nonzero(self, array: na.AbstractScalarArray):

            super().test_nonzero(array)

            result = np.nonzero(array)

            if not array.shape:
                assert result == dict()
                return

            expected = np.nonzero(array.ndarray)

            for i, ax in enumerate(array.axes):
                assert np.all(result[ax].ndarray == expected[i])
                assert len(result[ax].axes) == 1
                assert result[ax].axes[0] == array.axes_flattened

        @pytest.mark.parametrize('copy', [False, True])
        def test_nan_to_num(self, array: na.AbstractScalarArray, copy: bool):

            super().test_nan_to_num(array=array, copy=copy)

            if not copy and not isinstance(array, na.AbstractExplicitArray):
                with pytest.raises(ValueError, match="can't write to an array .*"):
                    np.nan_to_num(array, copy=copy)
                return

            result = np.nan_to_num(array, copy=copy)
            expected = np.nan_to_num(array.ndarray, copy=copy)

            if not copy:
                assert result is array

            assert np.all(result.ndarray == expected)

        @pytest.mark.parametrize('v', _scalar_arrays_2())
        @pytest.mark.parametrize('mode', ['full', 'same', 'valid'])
        def test_convolve(self, array: na.AbstractArray, v: na.AbstractArray, mode: str):
            super().test_convolve(array=array, v=v, mode=mode)

            shape_broadcasted = na.shape_broadcasted(array, v)

            if v is None:
                assert np.convolve(array, v, mode=mode) == na.ScalarArray(None)
                return

            if len(shape_broadcasted) > 1:
                with pytest.raises(ValueError, match=r"\'a\' and \'v\' must broadcast to .*"):
                    np.convolve(array, v, mode=mode)
                return

            result = np.convolve(array, v, mode=mode)
            result_expected = np.convolve(
                array.ndarray,
                v.ndarray if isinstance(v, na.AbstractArray) else v,
                mode=mode,
            )

            assert result.axes == tuple(shape_broadcasted.keys())
            assert np.all(result.ndarray == result_expected)

    class TestNamedArrayFunctions(
        AbstractTestAbstractScalar.TestNamedArrayFunctions
    ):
        class TestInterp(
            AbstractTestAbstractScalar.TestNamedArrayFunctions.TestInterp,
        ):
            def test_interp(
                    self,
                    array: na.AbstractScalarArray,
                    slope: float | na.AbstractArray,
            ):
                if np.issubdtype(array.dtype, bool):
                    return
                super().test_interp(array=array, slope=slope)

        @pytest.mark.parametrize(
            argnames="array_2",
            argvalues=_scalar_arrays_2()[:8],
        )
        @pytest.mark.parametrize(
            argnames="where, alpha",
            argvalues=[
                (
                    np._NoValue,
                    np._NoValue
                ),
                (
                    True,
                    na.linspace(0, 1, axis="x", num=_num_x, centers=True),
                ),
                (
                    na.linspace(0, 1, axis="x", num=_num_x) > 0.5,
                    0.5,
                )
            ]
        )
        class TestPltPlotLikeFunctions(
            AbstractTestAbstractScalar.TestNamedArrayFunctions.TestPltPlotLikeFunctions,
        ):
            pass

        @pytest.mark.parametrize(
            argnames="array_2",
            argvalues=_scalar_arrays_2()[:8],
        )
        @pytest.mark.parametrize(
            argnames="s,c,where",
            argvalues=[
                (
                    None,
                    None,
                    None,
                ),
                (
                    5,
                    na.ScalarArray(
                        ndarray=np.array([.5, .5, .5, 1]),
                        axes="rgba",
                    ),
                    na.linspace(0, 1, axis="x", num=_num_x) > 0.5,
                ),
                (
                    na.linspace(1, 2, axis="x", num=_num_x, endpoint=False, centers=True),
                    5,
                    True,
                )
            ],
        )
        class TestPltScatter(
            AbstractTestAbstractScalar.TestNamedArrayFunctions.TestPltScatter,
        ):
            pass

        @pytest.mark.parametrize("axis_x", ["x"])
        @pytest.mark.parametrize("axis_y", ["y"])
        class TestPltImshow:

            @pytest.mark.parametrize("axis_rgb", [None, "rgb"])
            def test_imshow(
                self,
                array: na.AbstractScalarArray,
                axis_x: str,
                axis_y: str,
                axis_rgb: None | str
            ):
                kwargs = dict(
                    X=array,
                    axis_x=axis_x,
                    axis_y=axis_y,
                    axis_rgb=axis_rgb,
                )

                if axis_x not in array.shape or axis_y not in array.shape:
                    with pytest.raises(ValueError):
                        na.plt.imshow(**kwargs)
                    return

                if axis_rgb is not None:
                    with pytest.raises(ValueError):
                        na.plt.imshow(**kwargs)
                    return

                result = na.plt.imshow(**kwargs)
                assert isinstance(result, na.ScalarArray)
                assert result.dtype == matplotlib.image.AxesImage

        class TestPltText:

            def test_plt_text(self, array):
                result = na.plt.text(
                    x=array,
                    y=array,
                    s="test label",
                )
                assert isinstance(result, na.ScalarArray)

        class TestPltBraceVertical:

            @pytest.mark.parametrize("kind", ["left", "right"])
            def test_plt_brace_vertical(
                self,
                array: na.AbstractScalarArray,
                kind: str,
            ):
                with astropy.visualization.quantity_support():
                    result = na.plt.brace_vertical(
                        x=array.value,
                        width=0.1,
                        ymin=0.1,
                        ymax=0.9,
                        kind=kind,
                    )
                assert isinstance(result, na.ScalarArray)

        @pytest.mark.parametrize(
            argnames="function",
            argvalues=[
                lambda x: a * x ** 3
                for a in [
                    2,
                    na.linspace(1, 2, num=6, axis="a"),
                    na.Cartesian2dVectorArray(2, 3),
                ]
            ]
        )
        class TestJacobian(
            AbstractTestAbstractScalar.TestNamedArrayFunctions.TestJacobian,
        ):
            pass

        @pytest.mark.parametrize(
            argnames="func",
            argvalues=[
                na.optimize.root_secant,
                na.optimize.root_newton,
            ],
        )
        @pytest.mark.parametrize(
            argnames="function",
            argvalues=[
                lambda x: np.square(na.value(x) - shift_horizontal) + shift_vertical
                for shift_horizontal in [
                    20,
                    na.linspace(19, 20, axis="c", num=6),
                ]
                for shift_vertical in [-1]
            ]
        )
        class TestOptimizeRoot(
            AbstractTestAbstractScalar.TestNamedArrayFunctions.TestOptimizeRoot,
        ):
            pass

        @pytest.mark.skip
        class TestOptimizeMinimum(
            AbstractTestAbstractScalar.TestNamedArrayFunctions.TestOptimizeMinimum,
        ):
            pass


@pytest.mark.parametrize('array', _scalar_arrays())
class TestScalarArray(
    AbstractTestAbstractScalarArray,
    tests.test_core.AbstractTestAbstractExplicitArray,
):

    @pytest.mark.parametrize(
        argnames="item",
        argvalues=[
            dict(y=0),
            dict(x=0, y=0),
            dict(y=slice(None)),
            dict(y=na.ScalarArrayRange(0, _num_y, axis='y')),
            dict(x=na.ScalarArrayRange(0, _num_x, axis='x'), y=na.ScalarArrayRange(0, _num_y, axis='y')),
            na.ScalarArray.ones(shape=dict(y=_num_y), dtype=bool),
        ],
    )
    @pytest.mark.parametrize(
        argnames="value",
        argvalues=[
            0,
            na.ScalarUniformRandomSample(-5, 5, dict(y=_num_y)),
        ]
    )
    def test__setitem__(
            self,
            array: na.ScalarArray,
            item: dict[str, int | slice | na.ScalarArray] | na.ScalarArray,
            value: float | na.ScalarArray
    ):
        super().test__setitem__(array=array, item=item, value=value)

    class TestNamedArrayFunctions(
        AbstractTestAbstractScalarArray.TestNamedArrayFunctions,
    ):
        class TestPltImshow(
            AbstractTestAbstractScalarArray.TestNamedArrayFunctions.TestPltImshow,
        ):
            @pytest.mark.parametrize("axis_rgb", ["rgb"])
            def test_imshow_rgb(
                    self,
                    array: na.AbstractScalarArray,
                    axis_x: str,
                    axis_y: str,
                    axis_rgb: None | str
            ):
                array = array.broadcast_to(array.shape | {axis_rgb: 3})

                kwargs = dict(
                    X=array,
                    axis_x=axis_x,
                    axis_y=axis_y,
                    axis_rgb=axis_rgb,
                )

                if axis_x not in array.shape or axis_y not in array.shape:
                    with pytest.raises(ValueError):
                        na.plt.imshow(**kwargs)
                    return

                result = na.plt.imshow(**kwargs)
                assert isinstance(result, na.ScalarArray)
                assert result.dtype == matplotlib.image.AxesImage


@pytest.mark.parametrize("type_array", [na.ScalarArray])
class TestScalarArrayCreation(
    tests.test_core.AbstractTestAbstractExplicitArrayCreation,
):

    @pytest.mark.parametrize("like", [None] + _scalar_arrays())
    class TestFromScalarArray(
        tests.test_core.AbstractTestAbstractExplicitArrayCreation.TestFromScalarArray
    ):

        def test_from_scalar_array(
                self,
                type_array: type[na.AbstractExplicitArray],
                a: None | float | u.Quantity | na.AbstractScalar,
                like: None | na.AbstractArray
        ):
            if isinstance(a, na.AbstractArray):
                if not isinstance(a, na.AbstractScalarArray):
                    with pytest.raises(TypeError):
                        type_array.from_scalar_array(a=a, like=like)
                    return

                super().test_from_scalar_array(
                    type_array=type_array,
                    a=a,
                    like=like,
                )

    @pytest.mark.parametrize('shape', [dict(x=3), dict(x=4, y=5)])
    @pytest.mark.parametrize('dtype', [int, float, complex])
    class TestNumpyArrayCreationFunctions:

        def test_empty(
                self,
                type_array: type[na.AbstractArray],
                shape: dict[str, int],
                dtype: None | type | np.dtype | str,
        ):
            result = type_array.empty(shape, dtype=dtype)
            assert result.shape == shape
            assert result.dtype == dtype

        def test_zeros(
                self,
                type_array: type[na.AbstractArray],
                shape: dict[str, int],
                dtype: None | type | np.dtype | str
        ):
            result = type_array.zeros(shape, dtype=dtype)
            assert result.shape == shape
            assert np.all(result == 0)
            assert result.dtype == dtype

        def test_ones(
                self,
                type_array: type[na.AbstractArray],
                shape: dict[str, int],
                dtype: None | type | np.dtype | str
        ):
            result = type_array.ones(shape, dtype=dtype)
            assert result.shape == shape
            assert np.all(result == 1)
            assert result.dtype == dtype


class AbstractTestAbstractImplicitScalarArray(
    AbstractTestAbstractScalarArray,
    tests.test_core.AbstractTestAbstractImplicitArray,
):
    pass


class AbstractTestAbstractScalarRandomSample(
    AbstractTestAbstractImplicitScalarArray,
    tests.test_core.AbstractTestAbstractRandomSample,
):
    pass


def _scalar_uniform_random_samples() -> list[na.ScalarUniformRandomSample]:
    starts = [
        0,
        na.ScalarArray(np.random.random(_num_x), axes=('x', )),
    ]
    stops = [
        10,
        na.ScalarArray(10 * np.random.random(_num_x) + 1, axes=('x', )),
    ]
    units = [None, u.mm]
    shapes_random = [dict(y=_num_y)]
    return [
        na.ScalarUniformRandomSample(
            start=start << unit if unit is not None else start,
            stop=stop << unit if unit is not None else stop,
            shape_random=shape_random,
        ) for start in starts for stop in stops for unit in units for shape_random in shapes_random
    ]


@pytest.mark.parametrize('array', _scalar_uniform_random_samples())
class TestScalarUniformRandomSample(
    AbstractTestAbstractScalarRandomSample,
    tests.test_core.AbstractTestAbstractUniformRandomSample,
):
    pass


def _scalar_normal_random_samples() -> list[na.ScalarNormalRandomSample]:
    centers = [
        0,
        na.ScalarArray(np.random.random(_num_x), axes=('x', )),
    ]
    widths = [
        10,
        na.ScalarArray(10 * np.random.random(_num_x) + 1, axes=('x', )),
    ]
    units = [None, u.mm]
    shapes_random = [dict(y=_num_y)]
    return [
        na.ScalarNormalRandomSample(
            center=center << unit if unit is not None else center,
            width=width << unit if unit is not None else width,
            shape_random=shape_random,
        ) for center in centers for width in widths for unit in units for shape_random in shapes_random
    ]


@pytest.mark.parametrize('array', _scalar_normal_random_samples())
class TestScalarNormalRandomSample(
    AbstractTestAbstractScalarRandomSample,
    tests.test_core.AbstractTestAbstractNormalRandomSample,
):
    pass


def _scalar_poisson_random_samples() -> list[na.ScalarPoissonRandomSample]:
    centers = [
        4,
        na.ScalarArray(np.random.randint(1, 5, _num_x), axes=('x', )),
    ]
    units = [None, u.mm]
    shapes_random = [dict(y=_num_y)]
    return [
        na.ScalarPoissonRandomSample(
            center=center << unit if unit is not None else center,
            shape_random=shape_random,
        ) for center in centers for unit in units for shape_random in shapes_random
    ]


@pytest.mark.parametrize('array', _scalar_poisson_random_samples())
class TestScalarPoissonRandomSample(
    AbstractTestAbstractScalarRandomSample,
    tests.test_core.AbstractTestAbstractPoissonRandomSample,
):
    pass


class AbstractTestAbstractParameterizedScalarArray(
    AbstractTestAbstractImplicitScalarArray,
    tests.test_core.AbstractTestAbstractParameterizedArray,
):
    pass


def _scalar_array_ranges() -> list[na.ScalarArrayRange]:
    starts = [0, ]
    steps = [1, 2.5, ]
    return [
        na.ScalarArrayRange(
            start=start,
            stop=start + step * _num_y,
            axis='y',
            step=step,
        ) for start in starts for step in steps
    ]


@pytest.mark.parametrize('array', _scalar_array_ranges())
class TestScalarArrayRange(
    AbstractTestAbstractParameterizedScalarArray,
    tests.test_core.AbstractTestAbstractArrayRange,
):
    pass


class AbstractTestAbstractScalarSpace(
    AbstractTestAbstractParameterizedScalarArray,
    tests.test_core.AbstractTestAbstractSpace,
):
    pass


def _scalar_linear_spaces():
    starts = [
        0,
        na.ScalarArray(np.random.random(_num_x), axes=('x', )),
    ]
    stops = [
        10,
        na.ScalarArray(10 * np.random.random(_num_x) + 1, axes=('x', )),
    ]
    units = [None, u.mm]
    nums = [_num_y]
    endpoints = [
        False,
        True,
    ]
    centers = [
        False,
        True,
    ]
    return [
        na.ScalarLinearSpace(
            start=start << unit if unit is not None else start,
            stop=stop << unit if unit is not None else stop,
            axis='y',
            num=num,
            endpoint=endpoint,
            centers=center,
        )
        for start in starts
        for stop in stops
        for unit in units
        for num in nums
        for endpoint, center in zip(endpoints, centers)
    ]


@pytest.mark.parametrize('array', _scalar_linear_spaces())
class TestScalarLinearSpace(
    AbstractTestAbstractScalarSpace,
    AbstractTestAbstractScalarArray,
    tests.test_core.AbstractTestAbstractLinearSpace,
    tests.test_core.AbstractTestAbstractArray,
):

    @pytest.mark.parametrize(
        argnames="axis",
        argvalues=[
            None,
            "y",
            "x",
        ]
    )
    def test_volume_cell(
        self,
        array: na.ScalarLinearSpace,
        axis: None | str | Sequence[str],
    ):
        super().test_volume_cell(array=array, axis=axis)

        axis_ = na.axis_normalized(array, axis)
        if len(axis_) != 1:
            with pytest.raises(ValueError):
                array.volume_cell(axis)
            return

        if not set(axis_).issubset(array.shape):
            with pytest.raises(ValueError):
                array.volume_cell(axis)
            return

        assert np.allclose(array.volume_cell(axis), array.explicit.volume_cell(axis))


def _scalar_stratified_random_spaces():
    starts = [
        0,
        na.ScalarArray(np.random.random(_num_x), axes=('x', )),
    ]
    stops = [
        10,
        na.ScalarArray(10 * np.random.random(_num_x) + 1, axes=('x', )),
    ]
    units = [None, u.mm]
    endpoints = [
        False,
        True,
    ]
    return [
        na.ScalarStratifiedRandomSpace(
            start=start << unit if unit is not None else start,
            stop=stop << unit if unit is not None else stop,
            axis='y',
            num=_num_y,
            endpoint=endpoint,
            seed=None
        ) for start in starts for stop in stops for unit in units for endpoint in endpoints
    ]


@pytest.mark.parametrize('array', _scalar_stratified_random_spaces())
class TestStratifiedRandomSpace(
    AbstractTestAbstractScalarSpace,
    tests.test_core.AbstractTestAbstractStratifiedRandomSpace,
):
    pass

def _scalar_logarithmic_spaces():
    start_exponents = [
        0,
        na.ScalarArray(np.random.random(_num_x), axes=('x', )),
    ]
    stop_exponents = [
        2,
        na.ScalarArray(np.random.random(_num_x) + 2, axes=('x', )),
    ]
    bases = [
        2,
    ]
    endpoints = [
        False,
        True,
    ]
    return [
        na.ScalarLogarithmicSpace(
            start_exponent=start,
            stop_exponent=stop,
            base=base,
            axis='y',
            num=_num_y,
            endpoint=endpoint,
        ) for start in start_exponents for stop in stop_exponents for base in bases for endpoint in endpoints
    ]


@pytest.mark.parametrize('array', _scalar_logarithmic_spaces())
class TestScalarLogarithmicSpace(
    AbstractTestAbstractScalarSpace,
    tests.test_core.AbstractTestAbstractLogarithmicSpace,
):
    pass

def _scalar_geometric_spaces():
    starts = [
        1,
        na.ScalarArray(np.random.random(_num_x), axes=('x', )),
    ]
    stops = [
        10,
        na.ScalarArray(10 * np.random.random(_num_x) + 1, axes=('x', )),
    ]
    units = [None, u.mm]
    endpoints = [
        False,
        True,
    ]
    return [
        na.ScalarGeometricSpace(
            start=start << unit if unit is not None else start,
            stop=stop << unit if unit is not None else stop,
            axis='y',
            num=_num_y,
            endpoint=endpoint
        ) for start in starts for stop in stops for unit in units for endpoint in endpoints
    ]


@pytest.mark.parametrize('array', _scalar_geometric_spaces())
class TestScalarGeometricSpace(
    AbstractTestAbstractScalarSpace,
    tests.test_core.AbstractTestAbstractGeometricSpace,
):
    pass



from __future__ import annotations
from typing import TypeVar, Generic, Iterator
from typing_extensions import Self
import abc
import dataclasses
import astropy.units as u
import named_arrays as na

__all__ = [
    "AbstractTransformation",
    "AbstractTranslation",
    "Translation",
    "AbstractLinearTransformation",
    "LinearTransformation",
    "AbstractAffineTransformation",
    "AffineTransformation",
    "AbstractTransformationList",
    "TransformationList",
]

VectorT = TypeVar("VectorT", bound="na.AbstractVectorArray")
MatrixT = TypeVar("MatrixT", bound="na.AbstractMatrixArray")
TransformationT = TypeVar("TransformationT", bound="AbstractTransformation")
LinearTransformationT = TypeVar("LinearTransformationT", bound="AbstractLinearTransformation")
TranslationT = TypeVar("TranslationT", bound="AbstractTranslation")


@dataclasses.dataclass(eq=False)
class AbstractTransformation(
    abc.ABC
):
    """
    An interface for an arbitrary vector transform
    """

    @property
    @abc.abstractmethod
    def shape(self) -> dict[str, int]:
        """The shape of the transformation."""

    @abc.abstractmethod
    def __call__(
            self,
            a: na.AbstractVectorArray,
    ) -> na.AbstractVectorArray:
        """
        apply the transformation to the given vector

        Parameters
        ----------
        a
            vector to transform
        """

    @property
    @abc.abstractmethod
    def inverse(self: Self) -> Self:
        """
        a new transformation that reverses the effect of this transformation
        """

    @abc.abstractmethod
    def __matmul__(
            self,
            other: AbstractTransformation,
    ) -> AbstractTransformation:
        """
        Compose multiple transformations into a single transformation.

        Parameters
        ----------
        other
            another transformation to compose with this one

        Examples
        --------
        Compose two transformations

        .. jupyter-execute::

            import astropy.units as u
            import named_arrays as na

            t1 = na.transformations.Translation(
                vector=na.Cartesian2dVectorArray(x=5) * u.mm,
            )
            t2 = na.transformations.LinearTransformation(
                matrix=na.Cartesian2dRotationArray(53 * u.deg)
            )

            t_composed = t1 @ t2
            t_composed

        use the composed transformation to transform a vector

        .. jupyter-execute::

            v = na.Cartesian3dVectorArray(1, 2, 3) * u.mm

            t_composed(v)

        transform the same vector by applying each transformation separately
        and note that it's the same result as using the composed transformation

        .. jupyter-execute::

            t1(t2(v))
        """


@dataclasses.dataclass(eq=False)
class IdentityTransformation(
    AbstractTransformation,
):
    """
    The identity transformation just returns its inputs.
    """

    @property
    def shape(self) -> dict[str, int]:
        return dict()

    def __call__(
        self,
        a: na.AbstractVectorArray,
    ) -> na.AbstractVectorArray:
        return a

    @property
    def inverse(self: Self) -> Self:
        return self

    def __matmul__(self, other: TransformationT) -> TransformationT:
        return other

    def __rmatmul__(self, other: TransformationT) -> TransformationT:
        return other


@dataclasses.dataclass(eq=False)
class AbstractTranslation(
    AbstractTransformation,
):
    @property
    @abc.abstractmethod
    def vector(self) -> na.AbstractVectorArray:
        """
        the vector representing the translation
        """

    @property
    def shape(self) -> dict[str, int]:
        return self.vector.shape

    def __call__(self, a: na.AbstractVectorArray) -> na.AbstractVectorArray:
        return a + self.vector

    @property
    def inverse(self: Self) -> Self:
        return Translation(vector=-self.vector)

    def __matmul__(self, other: AbstractTranslation) -> AbstractTranslation:
        if isinstance(other, AbstractTranslation):
            return Translation(
                vector=self.vector + other.vector
            )
        else:
            return NotImplemented


@dataclasses.dataclass(eq=False)
class Translation(
    AbstractTranslation,
    Generic[VectorT]
):
    """
    A translation-only vector transformation.

    Examples
    --------

    Translate a vector using a single transformation

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na

        vector = na.Cartesian3dVectorArray(
            x=12 * u.mm,
            y=12 * u.mm,
        )

        transformation = na.transformations.Translation(vector)

        square = na.Cartesian3dVectorArray(
            x=na.ScalarArray([-10, 10, 10, -10, -10] * u.mm, axes="vertex"),
            y=na.ScalarArray([-10, -10, 10, 10, -10] * u.mm, axes="vertex"),
        )

        square_transformed = transformation(square)

        with astropy.visualization.quantity_support():
            plt.figure();
            plt.gca().set_aspect("equal");
            na.plt.plot(square, label="original");
            na.plt.plot(square_transformed, label="translated");
            plt.legend();

    |

    Translate a vector using an array of transformations

    .. jupyter-execute::

        vector_2 = na.Cartesian3dVectorArray(
            x=na.ScalarArray([12, -12] * u.mm, axes="transform"),
            y=9 * u.mm,
        )

        transformation_2 = na.transformations.Translation(vector_2)

        square_transformed_2 = transformation_2(square)

        with astropy.visualization.quantity_support():
            plt.figure();
            plt.gca().set_aspect("equal");
            na.plt.plot(square, label="original");
            na.plt.plot(square_transformed_2, axis="vertex", label="translated");
            plt.legend();
    """
    vector: VectorT = dataclasses.MISSING


@dataclasses.dataclass(eq=False)
class Cartesian3dTranslation(
    AbstractTranslation
):
    x: na.ScalarLike = 0 * u.mm
    y: na.ScalarLike = 0 * u.mm
    z: na.ScalarLike = 0 * u.mm

    @property
    def vector(self) -> na.Cartesian3dVectorArray:
        return na.Cartesian3dVectorArray(self.x, self.y, self.z)


@dataclasses.dataclass(eq=False)
class AbstractLinearTransformation(
    AbstractTransformation,
):
    @property
    @abc.abstractmethod
    def matrix(self) -> na.AbstractMatrixArray:
        """
        the matrix representing the linear transformation
        """

    @property
    def shape(self) -> dict[str, int]:
        return self.matrix.shape

    def __call__(self, a: na.AbstractVectorArray) -> na.AbstractVectorArray:
        return self.matrix @ a

    @property
    def inverse(self: Self) -> Self:
        return LinearTransformation(matrix=self.matrix.inverse)

    def __matmul__(
            self,
            other: AbstractLinearTransformation | AbstractTranslation,
    ) -> LinearTransformation | AffineTransformation:
        if isinstance(other, AbstractLinearTransformation):
            return LinearTransformation(
                self.matrix @ other.matrix,
            )
        elif isinstance(other, AbstractTranslation):
            return AffineTransformation(
                transformation_linear=self,
                translation=Translation(self.matrix @ other.vector),
            )
        else:
            return NotImplemented

    def __rmatmul__(
            self,
            other: AbstractTranslation,
    ) -> AffineTransformation:
        if isinstance(other, AbstractTranslation):
            return AffineTransformation(
                transformation_linear=self,
                translation=other,
            )


@dataclasses.dataclass(eq=False)
class LinearTransformation(
    AbstractLinearTransformation,
    Generic[MatrixT],
):
    """
    A vector transformation represented by a matrix multiplication

    Examples
    --------

    Rotate a vector using a linear transformation

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na

        angle = 53 * u.deg
        matrix = na.Cartesian2dRotationMatrixArray(angle)

        transformation = na.transformations.LinearTransformation(matrix)

        square = na.Cartesian2dVectorArray(
            x=na.ScalarArray([-10, 10, 10, -10, -10] * u.mm, axes="vertex"),
            y=na.ScalarArray([-10, -10, 10, 10, -10] * u.mm, axes="vertex"),
        )

        square_transformed = transformation(square)

        with astropy.visualization.quantity_support():
            plt.figure();
            plt.gca().set_aspect("equal");
            na.plt.plot(square, label="original");
            na.plt.plot(square_transformed, label="rotated");
            plt.legend();

    |

    Rotate a vector using an array of transformations

    .. jupyter-execute::

        angle_2 = na.ScalarArray([30, 45] * u.deg, axes="transform")
        matrix_2 = na.Cartesian2dRotationMatrixArray(angle_2)

        transformation_2 = na.transformations.LinearTransformation(matrix_2)

        square_transformed_2 = transformation_2(square)

        with astropy.visualization.quantity_support():
            plt.figure();
            plt.gca().set_aspect("equal");
            na.plt.plot(square, label="original");
            na.plt.plot(square_transformed_2, axis="vertex", label="rotated");
            plt.legend();
    """
    matrix: MatrixT = dataclasses.MISSING


@dataclasses.dataclass(eq=False)
class AbstractCartesian3dRotation(
    AbstractLinearTransformation
):
    angle: na.ScalarLike = 0 * u.deg

    @classmethod
    @abc.abstractmethod
    def _matrix_type(cls) -> type[na.AbstractCartesian3dRotationMatrixArray]:
        """to be used by subclasses to specify which rotation matrix to use"""

    @property
    def matrix(self) -> na.AbstractCartesian3dRotationMatrixArray:
        return self._matrix_type()(self.angle)


@dataclasses.dataclass(eq=False)
class Cartesian3dRotationX(
    AbstractCartesian3dRotation
):
    def _matrix_type(cls):
        return na.Cartesian3dXRotationMatrixArray


@dataclasses.dataclass(eq=False)
class Cartesian3dRotationY(
    AbstractCartesian3dRotation
):
    def _matrix_type(cls):
        return na.Cartesian3dYRotationMatrixArray


@dataclasses.dataclass(eq=False)
class Cartesian3dRotationZ(
    AbstractCartesian3dRotation
):
    def _matrix_type(cls):
        return na.Cartesian3dZRotationMatrixArray


@dataclasses.dataclass(eq=False)
class AbstractAffineTransformation(
    AbstractTransformation,
):

    @property
    @abc.abstractmethod
    def transformation_linear(self) -> AbstractLinearTransformation:
        """
        The linear transformation component of this affine transformation
        """

    @property
    @abc.abstractmethod
    def translation(self) -> AbstractTranslation:
        """
        the translation component of this affine transformation
        """

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            self.transformation_linear.shape,
            self.translation.shape,
        )

    def __call__(self, a: na.AbstractVectorArray) -> na.AbstractVectorArray:
        return self.translation(self.transformation_linear(a))

    @property
    def inverse(self):
        transformation_linear = self.transformation_linear.inverse
        translation = self.translation.inverse
        return AffineTransformation(
            transformation_linear=transformation_linear,
            translation=Translation(transformation_linear(translation.vector)),
        )

    def __matmul__(
            self,
            other: AbstractAffineTransformation | AbstractLinearTransformation | AbstractTranslation
    ) -> AffineTransformation:
        transformation_linear = self.transformation_linear
        translation = self.translation
        if isinstance(other, AbstractAffineTransformation):
            return AffineTransformation(
                transformation_linear=transformation_linear @ other.transformation_linear,
                translation=Translation(transformation_linear.matrix @ other.translation.vector + translation.vector),
            )
        elif isinstance(other, AbstractLinearTransformation):
            return AffineTransformation(
                transformation_linear=transformation_linear @ other,
                translation=translation,
            )
        elif isinstance(other, AbstractTranslation):
            return AffineTransformation(
                transformation_linear=transformation_linear,
                translation=Translation(transformation_linear.matrix @ other.vector + translation.vector)
            )
        else:
            return NotImplemented

    def __rmatmul__(
            self,
            other: AbstractLinearTransformation | AbstractTranslation,
    ) -> AffineTransformation:
        if isinstance(other, AbstractLinearTransformation):
            return AffineTransformation(
                transformation_linear=other @ self.transformation_linear,
                translation=Translation(other.matrix @ self.translation.vector),
            )
        elif isinstance(other, AbstractTranslation):
            return AffineTransformation(
                transformation_linear=self.transformation_linear,
                translation=other @ self.translation,
            )
        else:
            return NotImplemented


@dataclasses.dataclass
class AffineTransformation(
    AbstractAffineTransformation,
    Generic[LinearTransformationT, TranslationT],
):
    transformation_linear: LinearTransformationT = dataclasses.MISSING
    translation: TranslationT = dataclasses.MISSING


@dataclasses.dataclass
class AbstractTransformationList(
    AbstractTransformation,
):
    @property
    @abc.abstractmethod
    def transformations(self) -> list[AbstractTransformation]:
        """
        the underlying list of transformations to compose together
        """

    @property
    @abc.abstractmethod
    def intrinsic(self) -> bool:
        """
        flag controlling whether the transformation should be applied to the
        coordinates or the coordinate system
        """

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(*[t.shape for t in self.transformations])

    def __iter__(self) -> Iterator[AbstractTransformation]:
        if self.intrinsic:
            return iter(self.transformations)
        else:
            return reversed(self.transformations)

    @property
    def composed(self) -> AbstractTransformation:
        transformations = list(self)
        result = IdentityTransformation()
        for t in transformations:
            result = t @ result
        return result

    def __call__(self, a: na.AbstractVectorArray) -> na.AbstractVectorArray:
        return self.composed(a)

    @property
    def inverse(self) -> AbstractTransformation:
        return self.composed.inverse

    def __matmul__(self, other: AbstractTransformation):
        return self.composed @ other

    def __rmatmul__(self, other: AbstractTransformation):
        return other @ self.composed


@dataclasses.dataclass
class TransformationList(
    AbstractTransformationList
):
    transformations: list[AbstractTransformation] = dataclasses.MISSING
    intrinsic: bool = True



from __future__ import annotations
from typing import Type
import abc
import dataclasses
import named_arrays as na

__all__ = [
    'AbstractMatrixArray',
    'AbstractExplicitMatrixArray',
    'AbstractImplicitMatrixArray',
    'AbstractMatrixRandomSample',
    'AbstractMatrixUniformRandomSample',
    'AbstractMatrixNormalRandomSample',
    'AbstractParameterizedMatrixArray',
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractMatrixArray(
    na.AbstractVectorArray,
):

    @property
    @abc.abstractmethod
    def type_vector(self) -> Type[na.AbstractExplicitVectorArray]:
        """
        The corresponding :class:`named_arrays.AbstractVectorArray` class
        """

    @property
    def components(self) -> dict[str, na.AbstractVectorArray]:
        return super().components

    @property
    def entries(self) -> dict[tuple[str, ...], na.ScalarLike]:
        rows = self.cartesian_nd.rows
        result = {}
        for r in rows:
            row = rows[r]
            if isinstance(row, na.AbstractArray):
                for c in row.components:
                    result[(r, c)] = row.components[c]
            else:
                result[(r,)] = row

        return result

    @property
    def rows(self) -> dict[str, na.AbstractVectorArray]:
        """
        Rows of the matrix, same as :attr:`components`
        """
        return self.components

    @property
    def is_consistent(self) -> bool:
        """
        Check if all the rows of the matrix have the same abstract type.
        """
        rows = self.cartesian_nd.rows
        rows_iter = iter(rows)
        try:
            row_prototype = rows[next(rows_iter)]
            return all(set(rows[r].components) == set(row_prototype.components) for r in rows_iter)
        except AttributeError:
            return False

    @property
    def is_square(self):
        """
        Check if this matrix has the same number of rows and columns.
        """
        if self.is_consistent:
            rows = self.cartesian_nd.rows
            num_rows = len(rows)
            return all(num_rows == len(rows[r].components) for r in rows)
        else:
            return False

    @property
    def matrix(self) -> AbstractMatrixArray:
        return self

    @property
    def row_prototype(self) -> na.AbstractVectorArray:
        """
        Return a vector of the same type as each row of the matrix.
        """
        rows = self.rows
        if not self.is_consistent:
            raise ValueError(
                f"matrix rows must all have the same abstract type, got "
                f"{tuple(rows[c].type_abstract for c in rows)}"
            )
        prototype_row = rows[next(iter(rows))]

        if isinstance(prototype_row, AbstractMatrixArray):
            prototype_row = prototype_row.row_prototype

        return prototype_row

    @property
    def column_prototype(self) -> na.AbstractMatrixArray:
        """
        Return a vector representing a column of the matrix with each component zeroed.
        """
        column_dict = {}
        for c in self.components:
            component = self.components[c]
            if isinstance(component, AbstractMatrixArray):
                column_dict[c] = component.column_prototype
            else:
                column_dict[c] = 0

        return self.type_vector.from_components(column_dict)

    def prototype_matrix(self, row: na.AbstractVectorArray = None):
        if row is None:
            row = self.row_prototype

        new_dict = {}
        components = self.column_prototype.matrix.components
        for c in components:
            component = components[c]
            if isinstance(component, AbstractMatrixArray):
                new_dict[c] = component.prototype_matrix(row)
            else:
                new_dict[c] = row

        return self.type_explicit.from_components(new_dict)

    @property
    def matrix_transpose(self):
        rows = self.rows

        if not self.is_consistent:
            raise ValueError(
                f"matrix rows must all have the same abstract type, got "
                f"{tuple(rows[c].type_abstract for c in rows)}"
            )

        new_row = self.column_prototype
        new_column = self.row_prototype.matrix
        # new_row = self.matrix_transpose.row_prototype

        prototype_matrix_transpose = new_column.prototype_matrix(new_row)

        result = prototype_matrix_transpose.cartesian_nd
        nd_rows = self.cartesian_nd.rows
        for r in nd_rows:
            for c in nd_rows[r].components:
                value = nd_rows[r].components[c]
                result.components[c].components[r] = value

        return prototype_matrix_transpose.from_cartesian_nd(result, like=prototype_matrix_transpose)

    @property
    @abc.abstractmethod
    def determinant(self) -> na.ScalarLike:
        """
        The determinant of this matrix
        """

    @property
    def inverse(self) -> AbstractMatrixArray:
        """
        The inverse of this matrix
        """
        explicit = self.explicit

        a = explicit.cartesian_nd

        if not a.is_square:
            raise ValueError("can only invert square matrices")

        rows = a.rows

        if len(rows) == 1:
            r_1, = rows.keys()
            row_1, = [r.components for r in rows.values()]
            c_1, = row_1.keys()
            a, = row_1.values()
            result = na.CartesianNdMatrixArray({
                c_1: na.CartesianNdVectorArray({
                    r_1: 1 / a
                })
            })

        elif len(rows) == 2:
            r_1, r_2 = rows.keys()
            row_1, row_2 = [r.components for r in rows.values()]
            c_1, c_2 = row_1.keys()
            a, b = row_1.values()
            c, d = row_2.values()
            det = a * d - b * c
            result = na.CartesianNdMatrixArray({
                c_1: na.CartesianNdVectorArray({
                    r_1: d / det,
                    r_2: -b / det,
                }),
                c_2: na.CartesianNdVectorArray({
                    r_1: -c / det,
                    r_2: a / det,
                }),
            })

        elif len(rows) == 3:
            r_1, r_2, r_3 = rows.keys()
            row_1, row_2, row_3 = [r.components for r in rows.values()]
            c_1, c_2, c_3 = row_1.keys()
            a, b, c = row_1.values()
            d, e, f = row_2.values()
            g, h, i = row_3.values()
            det = (a * e * i) + (b * f * g) + (c * d * h) - (c * e * g) - (b * d * i) - (a * f * h)
            result = na.CartesianNdMatrixArray({
                c_1: na.CartesianNdVectorArray({
                    r_1: (e * i - f * h) / det,
                    r_2: -(b * i - c * h) / det,
                    r_3: (b * f - c * e) / det,
                }),
                c_2: na.CartesianNdVectorArray({
                    r_1: -(d * i - f * g) / det,
                    r_2: (a * i - c * g) / det,
                    r_3: -(a * f - c * d) / det,
                }),
                c_3: na.CartesianNdVectorArray({
                    r_1: (d * h - e * g) / det,
                    r_2: -(a * h - b * g) / det,
                    r_3: (a * e - b * d) / det,
                }),
            })

        else:
            unit = na.unit(a, squeeze=False)
            value = na.value(a)
            value = na.stack([
                na.stack(list(row.components.values()), axis="_column")
                for row in value.rows.values()
            ], axis="_row")
            inverse = value.matrix_inverse(axis_rows="_row", axis_columns="_column")
            result = 1 / unit.matrix_transpose
            for i, r in enumerate(result.rows):
                row = result.rows[r].components
                for j, c in enumerate(row):
                    row[c] = inverse[dict(_row=i, _column=j)] * row[c]

        result = explicit.from_cartesian_nd(result, like=explicit.matrix_transpose)

        return result

    @property
    def cartesian_nd(self) -> na.AbstractCartesianNdMatrixArray:
        """
        Convert all cartesian vectors making up the matrix to instances of :class:`AbstractCartesianNdVectorArray`
        """
        components_new = dict()
        components = self.components
        for c in components:
            component = components[c]

            if isinstance(component, na.AbstractMatrixArray):
                component2 = component.cartesian_nd.components
                for c2 in component2:
                    components_new[f"{c}_{c2}"] = component2[c2]
            elif isinstance(component, na.AbstractVectorArray):
                components_new[c] = component.cartesian_nd
            else:
                components_new[c] = component

        return na.CartesianNdMatrixArray(components_new)

    def __array_matmul__(
            self,
            x1: na.ArrayLike,
            x2: na.ArrayLike,
            out: tuple[None | na.AbstractExplicitArray] = (None,),
            **kwargs,
    ) -> na.AbstractExplicitArray:

        if isinstance(x1, na.AbstractMatrixArray):
            components_x1 = x1.cartesian_nd.components

            if isinstance(x2, na.AbstractMatrixArray):
                if x1.row_prototype.cartesian_nd.components.keys() == x2.column_prototype.cartesian_nd.components.keys():

                    prototype_matrix = x1.prototype_matrix(x2.row_prototype)

                    x2 = x2.matrix_transpose
                    components_x2 = x2.cartesian_nd.components

                    result = dict()
                    for r in components_x1:
                        result[r] = na.CartesianNdVectorArray(
                            {c: components_x1[r] @ components_x2[c] for c in components_x2}
                        )
                    result = prototype_matrix.from_cartesian_nd(
                        na.CartesianNdMatrixArray(result),
                        like=prototype_matrix,
                    )
                else:
                    result = NotImplemented

            elif isinstance(x2, na.AbstractVectorArray):
                if x1.row_prototype.type_abstract == x2.type_abstract:
                    result_components = {r: components_x1[r] @ x2.cartesian_nd for r in components_x1}
                    result = x1.type_vector.from_cartesian_nd(
                        array=na.CartesianNdVectorArray(result_components),
                        like=x1.column_prototype,
                    )

                else:
                    result = NotImplemented
            else:
                result_components = {r: components_x1[r] @ x2 for r in components_x1}
                result = x1.type_explicit.from_cartesian_nd(na.CartesianNdMatrixArray(result_components), like=x1)

        else:
            if isinstance(x2, na.AbstractMatrixArray):

                x2 = x2.matrix_transpose
                components_x2 = x2.cartesian_nd.components

                if isinstance(x1, na.AbstractVectorArray):
                    component_dict = {c: x1.cartesian_nd @ components_x2[c] for c in components_x2}
                    result = x1.type_explicit.from_cartesian_nd(na.CartesianNdVectorArray(component_dict), like=x1)
                else:
                    component_dict = {c: x1 @ components_x2[c] for c in components_x2}
                    result = x2.type_explicit.from_cartesian_nd(na.CartesianNdMatrixArray(component_dict), like=x2)

            else:
                result = NotImplemented

        return result


@dataclasses.dataclass(eq=False, repr=False)
class AbstractExplicitMatrixArray(
    AbstractMatrixArray,
    na.AbstractExplicitVectorArray,
):

    @property
    def components(self) -> dict[str, na.AbstractVectorArray]:
        return self.__dict__

    @components.setter
    def components(self, value: dict[str, na.AbstractVectorArray]):
        self.__dict__ = value

    @classmethod
    def from_cartesian_nd(
        cls,
        array: na.CartesianNdMatrixArray,
        like: None | AbstractExplicitMatrixArray = None
    ) -> AbstractExplicitMatrixArray:

        if like is None:
            components_new = array.components

        else:
            nd_components = array.components
            components_new = {}
            components = like.components
            for c in components:

                component = components[c]
                if isinstance(component, na.AbstractVectorArray):
                    if isinstance(component, na.AbstractMatrixArray):
                        nd_key_mod = f"{c}_"
                        sub_dict = {k[len(nd_key_mod):]: v for k, v in nd_components.items() if k.startswith(nd_key_mod)}
                        components_new[c] = component.type_explicit.from_cartesian_nd(
                             na.CartesianNdMatrixArray(sub_dict),
                             like=component,
                         )
                    else:
                        components_new[c] = component.type_explicit.from_cartesian_nd(nd_components[c], like=component)
                else:
                    components_new[c] = nd_components[c]

        return cls.from_components(components_new)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImplicitMatrixArray(
    AbstractMatrixArray,
    na.AbstractImplicitVectorArray
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractMatrixRandomSample(
    AbstractImplicitMatrixArray,
    na.AbstractVectorRandomSample,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractMatrixUniformRandomSample(
    AbstractMatrixRandomSample,
    na.AbstractVectorUniformRandomSample,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractMatrixNormalRandomSample(
    AbstractMatrixRandomSample,
    na.AbstractVectorUniformRandomSample,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractParameterizedMatrixArray(
    AbstractImplicitMatrixArray,
    na.AbstractParameterizedVectorArray,
):
    pass



from __future__ import annotations
from typing import Type, Generic, TypeVar
from typing_extensions import Self
import abc
import dataclasses
import named_arrays as na

__all__ = [
    'AbstractTemporalVectorArray',
    'TemporalVectorArray',
    'AbstractImplicitTemporalVectorArray',
    'AbstractParameterizedTemporalVectorArray',
    'AbstractTemporalVectorSpace',
    'TemporalVectorLinearSpace',

]

TimeT = TypeVar("TimeT", bound=na.ArrayLike)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractTemporalVectorArray(
    na.AbstractCartesianVectorArray,
):

    @property
    @abc.abstractmethod
    def time(self) -> na.ArrayLike:
        """
        The temporal component of the vector.
        """

    @property
    def type_abstract(self) -> Type[na.AbstractArray]:
        return AbstractTemporalVectorArray

    @property
    def type_explicit(self) -> Type[na.AbstractExplicitArray]:
        return TemporalVectorArray

    @property
    def type_matrix(self) -> Type[na.TemporalMatrixArray]:
        return na.TemporalMatrixArray


@dataclasses.dataclass(eq=False, repr=False)
class TemporalVectorArray(
    AbstractTemporalVectorArray,
    na.AbstractExplicitCartesianVectorArray,
    Generic[TimeT],
):
    time: TimeT = 0

    @classmethod
    def from_scalar(
            cls: Type[Self],
            scalar: na.AbstractScalar,
            like: None | na.AbstractExplicitVectorArray = None,
    ) -> TemporalVectorArray:
        result = super().from_scalar(scalar, like=like)
        if result is not NotImplemented:
            return result
        return cls(time=scalar)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImplicitTemporalVectorArray(
    AbstractTemporalVectorArray,
    na.AbstractImplicitCartesianVectorArray,
):

    @property
    def time(self) -> na.ArrayLike:
        return self.explicit.time


@dataclasses.dataclass(eq=False, repr=False)
class AbstractParameterizedTemporalVectorArray(
    AbstractImplicitTemporalVectorArray,
    na.AbstractParameterizedVectorArray,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractTemporalVectorSpace(
    AbstractParameterizedTemporalVectorArray,
    na.AbstractVectorSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class TemporalVectorLinearSpace(
    AbstractTemporalVectorSpace,
    na.AbstractVectorLinearSpace,
):
    pass



from __future__ import annotations
from typing import Type, Generic, TypeVar
from typing_extensions import Self
import abc
import dataclasses
import named_arrays as na

__all__ = [
    "AbstractDirectionalVectorArray",
    "DirectionalVectorArray",
    "AbstractImplicitDirectionalVectorArray",
    "AbstractParameterizedDirectionalVectorArray",
    "AbstractDirectionalVectorSpace",
    "DirectionalVectorLinearSpace",
]

DirectionT = TypeVar("DirectionT", bound=na.ArrayLike)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractDirectionalVectorArray(
    na.AbstractCartesianVectorArray,
):

    @property
    @abc.abstractmethod
    def direction(self) -> na.ArrayLike:
        """
        The `direction` component of the vector.
        """

    @property
    def type_abstract(self) -> Type[na.AbstractArray]:
        return AbstractDirectionalVectorArray

    @property
    def type_explicit(self) -> Type[na.AbstractExplicitArray]:
        return DirectionalVectorArray

    @property
    def type_matrix(self) -> Type[na.DirectionalMatrixArray]:
        return na.DirectionalMatrixArray


@dataclasses.dataclass(eq=False, repr=False)
class DirectionalVectorArray(
    AbstractDirectionalVectorArray,
    na.AbstractExplicitCartesianVectorArray,
    Generic[DirectionT],
):
    direction: DirectionT = 0

    @classmethod
    def from_scalar(
        cls: Type[Self],
        scalar: na.AbstractScalar,
        like: None | na.AbstractExplicitVectorArray = None,
    ) -> DirectionalVectorArray:
        result = super().from_scalar(scalar, like=like)
        if result is not NotImplemented:
            return result
        return cls(direction=scalar)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImplicitDirectionalVectorArray(
    AbstractDirectionalVectorArray,
    na.AbstractImplicitCartesianVectorArray,
):

    @property
    def direction(self) -> na.ArrayLike:
        return self.explicit.direction


@dataclasses.dataclass(eq=False, repr=False)
class AbstractParameterizedDirectionalVectorArray(
    AbstractImplicitDirectionalVectorArray,
    na.AbstractParameterizedVectorArray,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractDirectionalVectorSpace(
    AbstractParameterizedDirectionalVectorArray,
    na.AbstractVectorSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class DirectionalVectorLinearSpace(
    AbstractDirectionalVectorSpace,
    na.AbstractVectorLinearSpace,
):
    pass
